import importlib.util as iutil
import logging
import os
from functools import partial
from pathlib import Path
from tempfile import gettempdir
from typing import List, Dict

import numpy as np

from aequilibrae import global_logger
from aequilibrae.paths.all_or_nothing import allOrNothing
from aequilibrae.paths.results import AssignmentResults
from aequilibrae.paths.traffic_class import TrafficClass
from aequilibrae.context import get_active_project
from ..utils import WorkerThread

try:
    from aequilibrae.paths.AoN import linear_combination, linear_combination_skims, aggregate_link_costs
    from aequilibrae.paths.AoN import triple_linear_combination, triple_linear_combination_skims
    from aequilibrae.paths.AoN import copy_one_dimension, copy_two_dimensions, copy_three_dimensions
    from aequilibrae.paths.AoN import sum_a_times_b_minus_c, linear_combination_1d
except ImportError as ie:
    global_logger.warning(f"Could not import procedures from the binary. {ie.args}")

import scipy

if int(scipy.__version__.split(".")[1]) >= 3:
    from scipy.optimize import root_scalar

    recent_scipy = True
else:
    from scipy.optimize import root as root_scalar

    recent_scipy = False
    global_logger.warning("Using older version of Scipy. For better performance, use Scipy >= 1.4")

if False:
    from aequilibrae.paths.traffic_assignment import TrafficAssignment

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import pyqtSignal as SIGNAL


class LinearApproximation(WorkerThread):
    if pyqt:
        equilibration = SIGNAL(object)
        assignment = SIGNAL(object)

    def __init__(self, assig_spec, algorithm, project=None) -> None:
        WorkerThread.__init__(self, None)

        self.logger = project.logger if project else logging.getLogger("aequilibrae")

        self.project_path = project.project_base_path if project else gettempdir()

        self.algorithm = algorithm
        self.rgap_target = assig_spec.rgap_target
        self.max_iter = assig_spec.max_iter
        self.cores = assig_spec.cores
        self.iteration_issue = []
        self.convergence_report = {"iteration": [], "rgap": [], "alpha": [], "warnings": []}
        if algorithm in ["cfw", "bfw"]:
            self.convergence_report["beta0"] = []
            self.convergence_report["beta1"] = []
            self.convergence_report["beta2"] = []

        self.assig = assig_spec  # type: TrafficAssignment

        if None in [
            assig_spec.classes,
            assig_spec.vdf,
            assig_spec.capacity_field,
            assig_spec.time_field,
            assig_spec.vdf_parameters,
        ]:
            all_par = "Traffic classes, VDF, VDF_parameters, capacity field & time_field"
            raise Exception(
                "Parameter missing. Setting the algorithm is the last thing to do "
                f"when assigning. Check if you have all of these: {all_par}"
            )

        self.traffic_classes = assig_spec.classes  # type: List[TrafficClass]
        self.num_classes = len(assig_spec.classes)

        self.cap_field = assig_spec.capacity_field
        self.time_field = assig_spec.time_field
        self.vdf = assig_spec.vdf
        self.vdf_parameters = assig_spec.vdf_parameters
        self.procedure_id = assig_spec.procedure_id

        self.iter = 0
        self.rgap = np.inf
        self.stepsize = 1.0
        self.conjugate_stepsize = 0.0
        self.fw_class_flow = 0
        # rgap can be a bit wiggly, specifying how many times we need to be below target rgap is a quick way to
        # ensure a better result. We might want to demand that the solution is that many consecutive times below.
        self.steps_below_needed_to_terminate = assig_spec.steps_below_needed_to_terminate
        self.steps_below = 0

        # if this is one, we do not have a new direction and will get stuck. Make it 1.
        self.conjugate_direction_max = 0.99999

        # if FW stepsize is zero, we set it to the corresponding MSA stepsize and then need to not make
        # the step direction conjugate to the previous direction.
        self.do_fw_step = False
        self.conjugate_failed = False
        self.do_conjugate_step = False

        # BFW specific stuff
        self.betas = np.array([1.0, 0.0, 0.0])

        # Instantiates the arrays that we will use over and over
        self.capacity = assig_spec.capacity
        self.free_flow_tt = assig_spec.free_flow_tt
        self.fw_total_flow = assig_spec.total_flow
        self.congested_time = assig_spec.congested_time
        self.vdf_der = np.array(assig_spec.congested_time, copy=True)
        self.congested_value = np.array(assig_spec.congested_time, copy=True)

        self.step_direction = {}  # type: Dict[AssignmentResults]
        self.previous_step_direction = {}  # type: Dict[AssignmentResults]
        self.pre_previous_step_direction = {}  # type: Dict[AssignmentResults]

        for c in self.traffic_classes:
            r = AssignmentResults()
            r.prepare(c.graph, c.matrix)
            self.step_direction[c.__id__] = r

        if self.algorithm in ["cfw", "bfw"]:
            for c in self.traffic_classes:
                r = AssignmentResults()
                r.prepare(c.graph, c.matrix)
                r.compact_link_loads = np.zeros([])
                r.compact_total_link_loads = np.zeros([])
                self.previous_step_direction[c.__id__] = r

                r = AssignmentResults()
                r.prepare(c.graph, c.matrix)
                r.compact_link_loads = np.zeros([])
                r.compact_total_link_loads = np.zeros([])
                self.step_direction[c.__id__] = r

                r = AssignmentResults()
                r.prepare(c.graph, c.matrix)
                r.compact_link_loads = np.zeros([])
                r.compact_total_link_loads = np.zeros([])
                self.pre_previous_step_direction[c.__id__] = r

    def calculate_conjugate_stepsize(self):
        self.vdf.apply_derivative(
            self.vdf_der, self.fw_total_flow, self.capacity, self.free_flow_tt, *self.vdf_parameters, self.cores
        )
        numerator = 0.0
        denominator = 0.0
        prev_dir_minus_current_sol = {}
        aon_minus_current_sol = {}
        aon_minus_prev_dir = {}

        for c in self.traffic_classes:
            stp_dir = self.step_direction[c.__id__]
            prev_dir_minus_current_sol[c.__id__] = np.sum(stp_dir.link_loads[:, :] - c.results.link_loads[:, :], axis=1)
            aon_minus_current_sol[c.__id__] = np.sum(
                c._aon_results.link_loads[:, :] - c.results.link_loads[:, :], axis=1
            )
            aon_minus_prev_dir[c.__id__] = np.sum(c._aon_results.link_loads[:, :] - stp_dir.link_loads[:, :], axis=1)

        for c_0 in self.traffic_classes:
            for c_1 in self.traffic_classes:
                numerator += prev_dir_minus_current_sol[c_0.__id__] * aon_minus_current_sol[c_1.__id__]
                denominator += prev_dir_minus_current_sol[c_0.__id__] * aon_minus_prev_dir[c_1.__id__]

        numerator = np.sum(numerator * self.vdf_der)
        denominator = np.sum(denominator * self.vdf_der)

        alpha = numerator / denominator
        if alpha < 0.0:
            self.conjugate_stepsize = 0.0
        elif alpha > self.conjugate_direction_max:
            self.conjugate_stepsize = self.conjugate_direction_max
        else:
            self.conjugate_stepsize = alpha

        # for reporting, we use a different convention, consistent with BFW: beta_0 corresponds to multiplier for AON;
        # in calculations we follow the conventions of our TRB paper.
        self.betas[0] = 1.0 - self.conjugate_stepsize
        self.betas[1] = self.conjugate_stepsize
        self.betas[2] = 0.0

    def calculate_biconjugate_direction(self):
        self.vdf.apply_derivative(
            self.vdf_der, self.fw_total_flow, self.capacity, self.free_flow_tt, *self.vdf_parameters, self.cores
        )
        mu_numerator = 0.0
        mu_denominator = 0.0
        nu_nom = 0.0
        nu_denom = 0.0

        w_ = {}
        x_ = {}
        y_ = {}
        z_ = {}

        for c in self.traffic_classes:
            sd = self.step_direction[c.__id__].link_loads[:, :]
            psd = self.previous_step_direction[c.__id__].link_loads[:, :]
            ll = c.results.link_loads[:, :]

            x_[c.__id__] = np.sum(sd * self.stepsize + psd * (1.0 - self.stepsize) - ll, axis=1)
            y_[c.__id__] = np.sum(c._aon_results.link_loads[:, :] - ll, axis=1)
            z_[c.__id__] = np.sum(sd - ll, axis=1)
            w_[c.__id__] = np.sum(psd - sd, axis=1)

        for c_0 in self.traffic_classes:
            for c_1 in self.traffic_classes:
                mu_numerator += x_[c_0.__id__] * y_[c_1.__id__]
                mu_denominator += x_[c_0.__id__] * w_[c_1.__id__]
                nu_nom += z_[c_0.__id__] * y_[c_1.__id__]
                nu_denom += z_[c_0.__id__] * z_[c_1.__id__]

        mu_numerator = np.sum(mu_numerator * self.vdf_der)
        mu_denominator = np.sum(mu_denominator * self.vdf_der)
        if mu_denominator == 0.0:
            mu = 0.0
        else:
            mu = -mu_numerator / mu_denominator
            mu = max(0.0, mu)

        nu_nom = np.sum(nu_nom * self.vdf_der)
        nu_denom = np.sum(nu_denom * self.vdf_der)
        if nu_denom == 0.0:
            nu = 0.0
        else:
            nu = -(nu_nom / nu_denom) + mu * self.stepsize / (1.0 - self.stepsize)
            nu = max(0.0, nu)

        self.betas[0] = 1.0 / (1.0 + nu + mu)
        self.betas[1] = nu * self.betas[0]
        self.betas[2] = mu * self.betas[0]

    def __calculate_step_direction(self):
        """Calculates step direction depending on the method"""
        sd_flows = []

        # 2nd iteration is a fw step. if the previous step replaced the aggregated
        # solution so far, we need to start anew.
        if self.iter == 2 or self.stepsize == 1.0 or self.do_fw_step or self.algorithm in ["msa", "frank-wolfe"]:
            self.do_fw_step = False
            self.do_conjugate_step = True
            self.conjugate_stepsize = 0.0
            for c in self.traffic_classes:
                aon_res = c._aon_results
                stp_dir_res = self.step_direction[c.__id__]
                copy_two_dimensions(stp_dir_res.link_loads, aon_res.link_loads, self.cores)
                stp_dir_res.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(stp_dir_res.skims.matrix_view, aon_res.skims.matrix_view, self.cores)
                sd_flows.append(aon_res.total_link_loads)

        # 3rd iteration is cfw. also, if we had to reset direction search we need a cfw step before bfw
        elif (self.iter == 3) or (self.do_conjugate_step) or (self.algorithm == "cfw"):
            self.do_conjugate_step = False
            self.calculate_conjugate_stepsize()
            for c in self.traffic_classes:
                sdr = self.step_direction[c.__id__]
                pre_previous = self.pre_previous_step_direction[c.__id__]
                copy_two_dimensions(pre_previous.link_loads, sdr.link_loads, self.cores)
                pre_previous.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(pre_previous.skims.matrix_view, sdr.skims.matrix_view, self.cores)

                linear_combination(
                    sdr.link_loads, sdr.link_loads, c._aon_results.link_loads, self.conjugate_stepsize, self.cores
                )

                if c.results.num_skims > 0:
                    linear_combination_skims(
                        sdr.skims.matrix_view,
                        sdr.skims.matrix_view,
                        c._aon_results.skims.matrix_view,
                        self.conjugate_stepsize,
                        self.cores,
                    )

                sdr.total_flows()
                sd_flows.append(sdr.total_link_loads)
        # biconjugate
        else:
            self.calculate_biconjugate_direction()
            # deep copy because we overwrite step_direction but need it on next iteration
            for c in self.traffic_classes:
                ppst = self.pre_previous_step_direction[c.__id__]  # type: AssignmentResults
                prev_stp_dir = self.previous_step_direction[c.__id__]  # type: AssignmentResults
                stp_dir = self.step_direction[c.__id__]  # type: AssignmentResults

                copy_two_dimensions(ppst.link_loads, stp_dir.link_loads, self.cores)
                ppst.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(ppst.skims.matrix_view, stp_dir.skims.matrix_view, self.cores)

                triple_linear_combination(
                    stp_dir.link_loads,
                    c._aon_results.link_loads,
                    stp_dir.link_loads,
                    prev_stp_dir.link_loads,
                    self.betas,
                    self.cores,
                )

                stp_dir.total_flows()
                if c.results.num_skims > 0:
                    triple_linear_combination_skims(
                        stp_dir.skims.matrix_view,
                        c._aon_results.skims.matrix_view,
                        stp_dir.skims.matrix_view,
                        prev_stp_dir.skims.matrix_view,
                        self.betas,
                        self.cores,
                    )

                sd_flows.append(np.sum(stp_dir.link_loads, axis=1))

                copy_two_dimensions(prev_stp_dir.link_loads, ppst.link_loads, self.cores)
                prev_stp_dir.total_flows()
                if c.results.num_skims > 0:
                    copy_three_dimensions(prev_stp_dir.skims.matrix_view, ppst.skims.matrix_view, self.cores)

        self.step_direction_flow = np.sum(sd_flows, axis=0)

    def __maybe_create_path_file_directories(self):
        path_base_dir = os.path.join(self.project_path, "path_files", self.procedure_id)
        for c in self.traffic_classes:
            if c._aon_results.save_path_file:
                c._aon_results.path_file_dir = os.path.join(
                    path_base_dir, f"iter{self.iter}", f"path_c{c.mode}_{c.__id__}"
                )
                Path(c._aon_results.path_file_dir).mkdir(parents=True, exist_ok=True)
                if self.iter == 1:  # save simplified graph correspondences, this could change after assignment
                    c.graph.save_compressed_correspondence(path_base_dir, c.mode, c.__id__)

    def doWork(self):
        self.execute()

    def execute(self):
        # We build the fixed cost field

        for c in self.traffic_classes:
            # Copying select link dictionary that maps name to its relevant matrices into the class' results
            c._aon_results._selected_links = c._selected_links
            c.results._selected_links = c._selected_links

            # Sizes the temporary objects used for the results
            c.results.prepare(c.graph, c.matrix)
            c._aon_results.prepare(c.graph, c.matrix)
            c.results.reset()

            # Prepares the fixed cost to be used
            if c.fixed_cost_field:
                # divide fixed cost by volume-dependent prefactor (vot) such that we don't have to do it for
                # each occurence in the objective funtion. TODO: Need to think about cost skims here, we do
                # not want this there I think
                v = c.graph.graph[c.fixed_cost_field].values[:]
                c.fixed_cost[c.graph.graph.__supernet_id__] = v * c.fc_multiplier / c.vot
                c.fixed_cost[np.isnan(c.fixed_cost)] = 0

        # TODO: Review how to eliminate this. It looks unnecessary
        # Just need to create some arrays for cost
        for c in self.traffic_classes:
            c.graph.set_graph(self.time_field)

        self.logger.info(f"{self.algorithm} Assignment STATS")
        self.logger.info("Iteration, RelativeGap, stepsize")
        for self.iter in range(1, self.max_iter + 1):
            self.iteration_issue = []
            if pyqt:
                self.equilibration.emit(["rgap", self.rgap])
                self.equilibration.emit(["iterations", self.iter])

            aon_flows = []

            self.__maybe_create_path_file_directories()

            for c in self.traffic_classes:  # type: TrafficClass
                # cost = c.fixed_cost / c.vot + self.congested_time #  now only once
                cost = c.fixed_cost + self.congested_time
                aggregate_link_costs(cost, c.graph.compact_cost, c.results.crosswalk)

                aon = allOrNothing(c.matrix, c.graph, c._aon_results)
                if pyqt:
                    aon.assignment.connect(self.signal_handler)
                aon.execute()
                c._aon_results.link_loads *= c.pce
                c._aon_results.total_flows()
                aon_flows.append(c._aon_results.total_link_loads)

            self.aon_total_flow = np.sum(aon_flows, axis=0)

            flows = []
            if self.iter == 1:
                for c in self.traffic_classes:
                    copy_two_dimensions(c.results.link_loads, c._aon_results.link_loads, self.cores)
                    c.results.total_flows()
                    if c.results.num_skims > 0:
                        copy_three_dimensions(c.results.skims.matrix_view, c._aon_results.skims.matrix_view, self.cores)

                    if c._selected_links:
                        for name, idx in c._aon_results._selected_links.items():
                            # Copy the temporary results into the final od matrix, referenced by link_set name
                            # The temp has an index associated with the link_set name
                            copy_three_dimensions(
                                c.results.select_link_od.matrix[name],  # matrix being written into
                                c._aon_results.temp_sl_od_matrix[idx, :, :, :],  # results after the iteration
                                self.cores,  # core count
                            )
                            copy_two_dimensions(
                                c.results.select_link_loading[name],  # ouput matrix
                                c._aon_results.temp_sl_link_loading[idx, :, :],  # matrix 1
                                self.cores,  # core count
                            )
                    flows.append(c.results.total_link_loads)

            else:
                self.__calculate_step_direction()
                self.calculate_stepsize()
                for c in self.traffic_classes:
                    stp_dir = self.step_direction[c.__id__]
                    cls_res = c.results
                    linear_combination(
                        cls_res.link_loads, stp_dir.link_loads, cls_res.link_loads, self.stepsize, self.cores
                    )

                    if cls_res.num_skims > 0:
                        linear_combination_skims(
                            cls_res.skims.matrix_view,
                            stp_dir.skims.matrix_view,
                            cls_res.skims.matrix_view,
                            self.stepsize,
                            self.cores,
                        )

                    if c._selected_links:
                        for name, idx in c._aon_results._selected_links.items():
                            # Copy the temporary results into the final od matrix, referenced by link_set name
                            # The temp flows have an index associated with the link_set name
                            linear_combination_skims(
                                c.results.select_link_od.matrix[name],  # output matrix
                                c._aon_results.temp_sl_od_matrix[idx, :, :, :],  # matrix 1
                                c.results.select_link_od.matrix[name],  # matrix 2 (previous iteration)
                                self.stepsize,  # stepsize
                                self.cores,  # core count
                            )

                            linear_combination(
                                c.results.select_link_loading[name],  # output matrix
                                c._aon_results.temp_sl_link_loading[idx, :, :],  # matrix 1
                                c.results.select_link_loading[name],  # matrix 2 (previous iteration)
                                self.stepsize,  # stepsize
                                self.cores,  # core count
                            )

                    cls_res.total_flows()
                    flows.append(cls_res.total_link_loads)

            self.fw_total_flow = np.sum(flows, axis=0)
            if self.algorithm == "all-or-nothing":
                break
            # Check convergence
            # This needs to be done with the current costs, and not the future ones
            converged = self.check_convergence() if self.iter > 1 else False

            self.vdf.apply_vdf(
                self.congested_time,
                self.fw_total_flow,
                self.capacity,
                self.free_flow_tt,
                *self.vdf_parameters,
                self.cores,
            )

            for c in self.traffic_classes:
                if self.time_field in c.graph.skim_fields:
                    k = c.graph.skim_fields.index(self.time_field)
                    aggregate_link_costs(self.congested_time[:], c.graph.compact_skims[:, k], c.results.crosswalk)

            self.convergence_report["iteration"].append(self.iter)
            self.convergence_report["rgap"].append(self.rgap)
            self.convergence_report["warnings"].append("; ".join(self.iteration_issue))
            self.convergence_report["alpha"].append(self.stepsize)

            if self.algorithm in ["cfw", "bfw"]:
                self.convergence_report["beta0"].append(self.betas[0])
                self.convergence_report["beta1"].append(self.betas[1])
                self.convergence_report["beta2"].append(self.betas[2])

            self.logger.info(f"{self.iter},{self.rgap},{self.stepsize}")
            if converged:
                self.steps_below += 1
                if self.steps_below >= self.steps_below_needed_to_terminate:
                    break
            else:
                self.steps_below = 0

            if self.iter < self.max_iter:
                for c in self.traffic_classes:
                    c._aon_results.reset()
                    if self.time_field not in c.graph.skim_fields:
                        continue
                    idx = c.graph.skim_fields.index(self.time_field)
                    c.graph.skims[:, idx] = self.congested_time[:]

        for c in self.traffic_classes:
            c.results.link_loads /= c.pce
            c.results.total_flows()

        if (self.rgap > self.rgap_target) and (self.algorithm != "all-or-nothing"):
            self.logger.error(f"Desired RGap of {self.rgap_target} was NOT reached")
        self.logger.info(f"{self.algorithm} Assignment finished. {self.iter} iterations and {self.rgap} final gap")
        if pyqt:
            self.equilibration.emit(["rgap", self.rgap])
            self.equilibration.emit(["iterations", self.iter])
            self.equilibration.emit(["finished_threaded_procedure"])

    def __derivative_of_objective_stepsize_dependent(self, stepsize, const_term):
        """The stepsize-dependent part of the derivative of the objective function. If fixed costs are defined,
        the corresponding contribution needs to be passed in"""
        x = np.zeros_like(self.fw_total_flow)
        linear_combination_1d(x, self.step_direction_flow, self.fw_total_flow, stepsize, self.cores)
        # x = self.fw_total_flow + stepsize * (self.step_direction_flow - self.fw_total_flow)
        self.vdf.apply_vdf(self.congested_value, x, self.capacity, self.free_flow_tt, *self.vdf_parameters, self.cores)
        link_cost_term = sum_a_times_b_minus_c(
            self.congested_value, self.step_direction_flow, self.fw_total_flow, self.cores
        )
        return link_cost_term + const_term

    def __derivative_of_objective_stepsize_independent(self):
        """The part of the derivative of the objective function that does not dependent on stepsize. Non-zero
        only for fixed cost contributions."""
        class_specific_term = 0.0
        for c in self.traffic_classes:
            # fixed cost is scaled by vot
            class_link_costs = sum_a_times_b_minus_c(
                c.fixed_cost, self.step_direction[c.__id__].link_loads[:, 0], c.results.link_loads[:, 0], self.cores
            )
            class_specific_term += class_link_costs
        return class_specific_term

    def calculate_stepsize(self):
        """Calculate optimal stepsize in descent direction"""
        if self.algorithm == "msa":
            self.stepsize = 1.0 / self.iter
            return

        class_specific_term = self.__derivative_of_objective_stepsize_independent()
        derivative_of_objective = partial(
            self.__derivative_of_objective_stepsize_dependent, const_term=class_specific_term
        )

        x_tol = max(min(1e-6, self.rgap * 1e-5), 1e-12)

        try:
            if recent_scipy:
                min_res = root_scalar(derivative_of_objective, bracket=[0, 1], xtol=x_tol)
                self.stepsize = min_res.root
                if not min_res.converged:
                    self.logger.warning("Descent direction stepsize finder has not converged")
            else:
                min_res = root_scalar(derivative_of_objective, 1 / self.iter, xtol=x_tol)
                if not min_res.success:
                    self.logger.warning("Descent direction stepsize finder has not converged")
                self.stepsize = min_res.x[0]
                if self.stepsize <= 0.0 or self.stepsize >= 1.0:
                    raise ValueError("wrong root")

            self.conjugate_failed = False

        except ValueError as e:
            # We can have iterations where the objective function is not *strictly* convex, but the scipy method cannot deal
            # with this. Stepsize is then either given by 1 or 0, depending on where the objective function is smaller.
            # However, using zero would mean the overall solution would not get updated, and therefore we assert the stepsize
            # in order to add a small fraction of the AoN. A heuristic value equal to the corresponding MSA step size
            # seems to work well in practice.
            if self.algorithm == "bfw":
                self.betas.fill(-1)
            if derivative_of_objective(0.0) < derivative_of_objective(1.0):
                if self.algorithm == "frank-wolfe" or self.conjugate_failed:
                    tiny_step = 1e-2 / self.iter  # use a fraction of the MSA stepsize. We observe that using 1e-4
                    # works well in practice, however for a large number of iterations this might be too much so
                    # use this heuristic instead.
                    self.logger.warning(f"# Alert: Adding {tiny_step} as step size to make it non-zero. {e.args}")
                    self.stepsize = tiny_step
                else:
                    self.stepsize = 0.0
                    # need to reset conjugate / bi-conjugate direction search
                    self.do_fw_step = True
                    self.conjugate_failed = True
                    self.iteration_issue.append("Found bad conjugate direction step. Performing FW search. {e.args}")
                    # By doing it recursively, we avoid doing the same AoN again
                    self.__calculate_step_direction()
                    self.calculate_stepsize()

            else:
                # Do we want to keep some of the old solution, or just throw away everything?
                self.stepsize = 1.0

        assert 0 <= self.stepsize <= 1.0

    def check_convergence(self):
        """Calculate relative gap and return True if it is smaller than desired precision"""
        aon_cost = np.sum(self.congested_time * self.aon_total_flow)
        current_cost = np.sum(self.congested_time * self.fw_total_flow)
        self.rgap = abs(current_cost - aon_cost) / current_cost
        if self.rgap_target >= self.rgap:
            return True
        return False

    def signal_handler(self, val):
        if pyqt:
            self.assignment.emit(val)



import random
from sqlite3 import Connection
from shapely.geometry import Point, MultiPolygon
from .network.safe_class import SafeClass
from .network.connector_creation import connector_creation


class Zone(SafeClass):
    """Single zone object that can be queried and manipulated in memory"""

    def __init__(self, dataset: dict, zoning):
        self.geometry = MultiPolygon()
        self.zone_id = -1
        super().__init__(dataset, zoning.project)
        self.__zoning = zoning
        self.conn = zoning.conn  # type: Connection
        self.__new = dataset["geometry"] is None
        self.__network_links = zoning.network.links
        self.__network_nodes = zoning.network.nodes

    def delete(self):
        """Removes the zone from the database"""
        conn = self._project.connect()
        curr = conn.cursor()
        curr.execute(f'DELETE FROM zones where zone_id="{self.zone_id}"')
        conn.commit()
        conn.close()
        self.__zoning._remove_zone(self.zone_id)
        del self

    def save(self):
        """Saves/Updates the zone data to the database"""

        if self.zone_id != self.__original__["zone_id"]:
            raise ValueError("One cannot change the zone_id")

        conn = self._project.connect()
        curr = conn.cursor()

        curr.execute(f'select count(*) from zones where zone_id="{self.zone_id}"')
        if curr.fetchone()[0] == 0:
            data = [self.zone_id, self.geometry.wkb]
            curr.execute("Insert into zones (zone_id, geometry) values(?, ST_Multi(GeomFromWKB(?, 4326)))", data)

        for key, value in self.__dict__.items():
            if key != "zone_id" and key in self.__original__:
                v_old = self.__original__.get(key, None)
                if value != v_old and value is not None:
                    self.__original__[key] = value
                    if key == "geometry":
                        sql = "update 'zones' set geometry=ST_Multi(GeomFromWKB(?, 4326)) where zone_id=?"
                        curr.execute(sql, [value.wkb, self.zone_id])
                    else:
                        curr.execute(f"update 'zones' set '{key}'=? where zone_id=?", [value, self.zone_id])
        conn.commit()
        conn.close()

    def add_centroid(self, point: Point, robust=True) -> None:
        """Adds a centroid to the network file

        :Arguments:
            **point** (:obj:`Point`): Shapely Point corresponding to the desired centroid position.
            If None, uses the geometric center of the zone
            **robust** (:obj:`Bool`, Optional): Moves the centroid location around to avoid node conflict.
            Defaults to True.
        """

        # This is VERY small in real-world terms (between zero and 11cm)
        shift = 0.000001

        curr = self.conn.cursor()

        curr.execute("select count(*) from nodes where node_id=?", [self.zone_id])
        if curr.fetchone()[0] > 0:
            self._project.logger.warning("Centroid already exists. Failed to create it")
            return

        sql = "INSERT into nodes (node_id, is_centroid, geometry) VALUES(?,1,GeomFromWKB(?, ?));"

        if point is None:
            point = self.geometry.centroid

        if robust:
            check_sql = """SELECT count(*) FROM nodes
                             WHERE  nodes.geometry = GeomFromWKB(?, 4326) AND
                          nodes.ROWID IN (
                           SELECT ROWID FROM SpatialIndex WHERE f_table_name = 'nodes' AND
                           search_frame = GeomFromWKB(?, 4326))
                       """

            test_list = self.conn.execute(check_sql, [point.wkb, point.wkb]).fetchone()
            while sum(test_list):
                test_list = self.conn.execute(check_sql, [point.wkb, point.wkb]).fetchone()
                point = Point(point.x + random.random() * shift, point.y + random.random() * shift)

        data = [self.zone_id, point.wkb, self.__srid__]
        self.conn.execute(sql, data)
        self.conn.commit()

    def connect_mode(self, mode_id: str, link_types="", connectors=1) -> None:
        """Adds centroid connectors for the desired mode to the network file

        Centroid connectors are created by connecting the zone centroid to one or more nodes selected from
        all those that satisfy the mode and link_types criteria and are inside the zone.

        The selection of the nodes that will be connected is done simply by computing running the
        `KMeans2 <https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.vq.kmeans2.html>`_
        clustering algorithm from SciPy and selecting the nodes closest to each cluster centroid.

        When there are no node candidates inside the zone, the search area is progressively expanded until
        at least one candidate is found.

        If fewer candidates than required connectors are found, all candidates are connected.

        :Arguments:
            **mode_id** (:obj:`str`): Mode ID we are trying to connect

            **link_types** (:obj:`str`, `Optional`): String with all the link type IDs that can be considered.
            eg: yCdR. Defaults to ALL link types

            **connectors** (:obj:`int`, `Optional`): Number of connectors to add. Defaults to 1
        """
        connector_creation(
            self.geometry,
            zone_id=self.zone_id,
            srid=self.__srid__,
            mode_id=mode_id,
            link_types=link_types,
            connectors=connectors,
            network=self._project.network,
        )

    def disconnect_mode(self, mode_id: str) -> None:
        """Removes centroid connectors for the desired mode from the network file

        :Arguments:
            **mode_id** (:obj:`str`): Mode ID we are trying to disconnect from this zone
        """

        curr = self.conn.cursor()
        data = [self.zone_id, mode_id]
        curr.execute("Delete from links where a_node=? and modes=?", data)
        row_count = curr.rowcount

        data = [mode_id, self.zone_id, mode_id]
        curr.execute('Update links set modes = replace(modes, ?, "") where a_node=? and instr(modes,?) > 0', data)
        row_count += curr.rowcount

        if row_count:
            self._project.logger.warning(f"Deleted {row_count} connectors for mode {mode_id} for zone {self.zone_id}")
        else:
            self._project.warning("No centroid connectors for this mode")
        self.conn.commit()



from sqlite3 import Connection
import shapely.wkb
import pandas as pd


class DataLoader:
    def __init__(self, conn: Connection, table_name: str):
        self.conn = conn
        self.curr = conn.cursor()
        self.table_name = table_name

    def load_table(self) -> pd.DataFrame:
        fields, _, geo_field = self.__find_table_fields()
        fields = [f'"{x}"' for x in fields]
        if geo_field is not None:
            fields.append('ST_AsBinary("geometry") geometry')
        keys = ",".join(fields)
        df = pd.read_sql_query(f"select {keys} from '{self.table_name}'", self.conn)
        df.geometry = df.geometry.apply(shapely.wkb.loads)
        return df

    def __find_table_fields(self):
        geotypes = ["LINESTRING", "POINT", "POLYGON", "MULTIPOLYGON"]
        self.curr.execute(f"pragma table_info({self.table_name})")
        structure = self.curr.fetchall()
        fields = [x[1].lower() for x in structure]
        geotype = geo_field = None
        for x in structure:
            if x[2].upper() in geotypes:
                geotype = x[2]
                geo_field = x[1]
                break
        if geo_field is not None:
            fields = [x for x in fields if x != geo_field.lower()]

        return fields, geotype, geo_field



import os
import uuid
from os.path import join, isfile
from shutil import rmtree
from tempfile import gettempdir
from unittest import TestCase

import numpy as np

from aequilibrae.paths import skimming_single_origin
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
from aequilibrae.paths.network_skimming import NetworkSkimming
from aequilibrae.paths.results import SkimResults
from aequilibrae.utils.create_example import create_example


class TestNetwork_skimming(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)

        self.project = create_example(self.proj_dir)
        self.network = self.project.network
        self.curr = self.project.conn.cursor()

    def tearDown(self) -> None:
        del self.curr
        try:
            rmtree(self.proj_dir)
        except Exception as e:
            print(f"Failed to remove at {e.args}")

    def test_network_skimming(self):
        self.network.build_graphs()
        graph = self.network.graphs["c"]
        graph.set_graph(cost_field="distance")
        graph.set_skimming("distance")
        graph.set_blocked_centroid_flows(False)

        # skimming results
        res = SkimResults()
        res.prepare(graph)
        aux_res = MultiThreadedNetworkSkimming()
        aux_res.prepare(graph, res)
        _ = skimming_single_origin(12, graph, res, aux_res, 0)

        skm = NetworkSkimming(graph)
        skm.execute()

        tot = np.nanmax(skm.results.skims.distance[:, :])
        if tot > np.sum(graph.cost):
            self.fail("Skimming was not successful. At least one np.inf returned.")

        if skm.report:
            self.fail("Skimming returned an error:" + str(skm.report))

        fn = "test_Skimming"
        skm.save_to_project(fn, format="omx")
        matrix_dir = join(self.proj_dir, "matrices")

        if not isfile(join(matrix_dir, f"{fn}.omx")):
            self.fail("Did not save project to project")

        matrices = self.project.matrices
        mat = matrices.get_record(fn)
        self.assertEqual(mat.name, fn, "Matrix record name saved wrong")
        self.assertEqual(mat.file_name, f"{fn}.omx", "matrix file_name saved  wrong")
        self.assertEqual(mat.cores, 1, "matrix saved number of matrix cores wrong")
        self.assertEqual(mat.procedure, "Network skimming", "Matrix saved wrong procedure name")
        self.assertEqual(mat.procedure_id, skm.procedure_id, "Procedure ID saved  wrong")
        self.assertEqual(mat.timestamp, skm.procedure_date, "Procedure ID saved  wrong")
        self.project.close()

    def test_network_skimming_no_project(self):
        self.network.build_graphs()
        graph = self.network.graphs["c"]
        graph.set_graph(cost_field="distance")
        graph.set_skimming("distance")
        graph.set_blocked_centroid_flows(False)

        self.project.close()
        # skimming results
        res = SkimResults()
        res.prepare(graph)
        aux_res = MultiThreadedNetworkSkimming()
        aux_res.prepare(graph, res)
        _ = skimming_single_origin(12, graph, res, aux_res, 0)

        skm = NetworkSkimming(graph)
        skm.execute()

        tot = np.nanmax(skm.results.skims.distance[:, :])
        if tot > np.sum(graph.cost):
            self.fail("Skimming was not successful. At least one np.inf returned.")

        if skm.report:
            self.fail("Skimming returned an error:" + str(skm.report))



import sys
import threading
import importlib.util as iutil
from uuid import uuid4
from multiprocessing.dummy import Pool as ThreadPool
from datetime import datetime
from aequilibrae.context import get_active_project
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
from aequilibrae.paths.results.skim_results import SkimResults
from aequilibrae.utils import WorkerThread
from aequilibrae import global_logger

try:
    from aequilibrae.paths.AoN import skimming_single_origin
except ImportError as ie:
    global_logger.warning(f"Could not import procedures from the binary. {ie.args}")

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import pyqtSignal

spec = iutil.find_spec("openmatrix")
has_omx = spec is not None

sys.dont_write_bytecode = True


class NetworkSkimming(WorkerThread):
    """

    .. code-block:: python

        >>> from aequilibrae import Project
        >>> from aequilibrae.paths.network_skimming import NetworkSkimming

        >>> project = Project.from_path("/tmp/test_project")

        >>> network = project.network
        >>> network.build_graphs()

        >>> graph = network.graphs['c']
        >>> graph.set_graph(cost_field="distance")
        >>> graph.set_skimming("distance")

        >>> skm = NetworkSkimming(graph)
        >>> skm.execute()

        # The skim report (if any error generated) is available here
        >>> skm.report
        []

        # To access the skim matrix directly from its temporary file
        >>> matrix = skm.results.skims

        # Or you can save the results to disk
        >>> skm.save_to_project('/tmp/skimming result.omx')

        # Or specify the AequilibraE's matrix file format
        >>> skm.save_to_project('skimming result', 'aem')

        >>> project.close()
    """

    if pyqt:
        skimming = pyqtSignal(object)

    def __init__(self, graph, origins=None, project=None):
        WorkerThread.__init__(self, None)
        self.project = project
        self.origins = origins
        self.graph = graph
        self.results = SkimResults()
        self.aux_res = MultiThreadedNetworkSkimming()
        self.report = []
        self.procedure_id = ""
        self.procedure_date = ""
        self.cumulative = 0

    def doWork(self):
        self.execute()

    def execute(self):
        """Runs the skimming process as specified in the graph"""
        if pyqt:
            self.skimming.emit(["zones finalized", 0])

        self.results.prepare(self.graph)
        self.aux_res = MultiThreadedNetworkSkimming()
        self.aux_res.prepare(self.graph, self.results)

        pool = ThreadPool(self.results.cores)
        all_threads = {"count": 0}
        for orig in list(self.graph.centroids):
            i = int(self.graph.nodes_to_indices[orig])
            if i >= self.graph.nodes_to_indices.shape[0]:
                self.report.append(f"Centroid {orig} is beyond the domain of the graph")
            elif self.graph.fs[int(i)] == self.graph.fs[int(i) + 1]:
                self.report.append(f"Centroid {orig} does not exist in the graph")
            else:
                pool.apply_async(self.__func_skim_thread, args=(orig, all_threads))
        pool.close()
        pool.join()
        self.aux_res = None
        self.procedure_id = uuid4().hex
        self.procedure_date = str(datetime.today())

        if pyqt:
            self.skimming.emit(["text skimming", "Saving Outputs"])
            self.skimming.emit(["finished_threaded_procedure", None])

    def save_to_project(self, name: str, format="omx", project=None) -> None:
        """Saves skim results to the project folder and creates record in the database

        :Arguments:
            **name** (:obj:`str`): Name of the matrix. Same value for matrix record name and file (plus extension)
            **format** (:obj:`str`, `Optional`): File format ('aem' or 'omx'). Default is 'omx'
            **project** (:obj:`Project`, `Optional`): Project we want to save the results to. Defaults to the active project
        """

        file_name = f"{name}.{format.lower()}"
        if not project:
            project = self.project or get_active_project()
        mats = project.matrices
        record = mats.new_record(name, file_name, self.results.skims)
        record.procedure_id = self.procedure_id
        record.timestamp = self.procedure_date
        record.procedure = "Network skimming"
        record.save()

    def __func_skim_thread(self, origin, all_threads):
        if threading.get_ident() in all_threads:
            th = all_threads[threading.get_ident()]
        else:
            all_threads[threading.get_ident()] = all_threads["count"]
            th = all_threads["count"]
            all_threads["count"] += 1
        x = skimming_single_origin(origin, self.graph, self.results, self.aux_res, th)
        self.cumulative += 1
        if x != origin:
            self.report.append(x)
        if pyqt:
            self.skimming.emit(["zones finalized", self.cumulative])
            txt = str(self.cumulative) + " / " + str(self.matrix.zones)
            self.skimming.emit(["text skimming", txt])



from random import randint, choice, uniform
from shapely.geometry import LineString
import pytest
from aequilibrae.project import Project
from aequilibrae.transit import Transit
from aequilibrae.project.database_connection import database_connection
from aequilibrae.transit.functions.get_srid import get_srid

from aequilibrae.transit.transit_elements import Stop
from tests.aequilibrae.transit.random_word import randomword


@pytest.fixture
def data():
    return {
        "stop_id": randint(0, 400000000),
        "stop_code": randomword(randint(0, 40)),
        "stop_name": randomword(randint(0, 40)),
        "stop_desc": randomword(randint(0, 40)),
        "stop_lat": uniform(0, 30000),
        "stop_lon": uniform(0, 30000),
        "stop_street": randomword(randint(0, 40)),
        "zone_id": randomword(randint(0, 40)),
        "stop_url": randomword(randint(0, 40)),
        "location_type": choice((0, 1)),
        "parent_station": randomword(randint(0, 40)),
        "stop_timezone": randint(0, 2000),
    }


def test__populate(data):
    s = Stop(1, tuple(data.values()), list(data.keys()))
    xy = (s.geo.x, s.geo.y)
    assert xy == (data["stop_lon"], data["stop_lat"]), "Stop built geo wrongly"
    data["stop"] = data["stop_id"]
    data["stop_id"] = s.stop_id
    data["zone"] = data.pop("zone_id")

    for key, val in s.__dict__.items():
        if key in data:
            assert val == data[key], "Stop population with record failed"

    new_data = {key: val for key, val in data.items()}
    new_data[randomword(randint(1, 15))] = randomword(randint(1, 20))
    with pytest.raises(KeyError):
        s = Stop(1, tuple(new_data.values()), list(new_data.keys()))


def test_save_to_database(data, transit_conn):
    line = LineString([[-23.59, -46.64], [-23.43, -46.50]]).wkb
    tlink_id = randint(10000, 200000044)
    s = Stop(1, tuple(data.values()), list(data.keys()))
    s.link = link = randint(1, 30000)
    s.dir = direc = choice((0, 1))
    s.agency = randint(5, 100000)
    s.route_type = randint(0, 13)
    s.srid = get_srid()
    s.get_node_id()
    s.save_to_database(transit_conn, commit=True)

    sql_tl = """Insert into route_links ("transit_link", "pattern_id", "seq", "from_stop", "to_stop", "distance", "geometry")
                VALUES(?, ?, ?, ?, ?, ?, GeomFromWKB(?, 4326));"""
    transit_conn.execute(sql_tl, [tlink_id, randint(1, 1000000000), randint(1, 10), s.stop_id, s.stop_id + 1, 0, line])

    qry = transit_conn.execute(
        "Select agency_id, link, dir, description, street from stops where stop=?", [data["stop_id"]]
    ).fetchone()
    result = [x for x in qry]
    expected = [s.agency_id, link, direc, data["stop_desc"], data["stop_street"]]
    assert result == expected, "Saving Stop to the database failed"



from os.path import isfile, join


def test_new_gtfs_builder(create_gtfs_project, create_path):
    transit = create_gtfs_project.new_gtfs_builder(
        agency="LISERCO, LISANCO, LINCOSUR",
        file_path=join(create_path, "gtfs_coquimbo.zip"),
    )

    assert str(type(transit)) == "<class 'aequilibrae.transit.lib_gtfs.GTFSRouteSystemBuilder'>"


def test___create_transit_database(create_gtfs_project):
    assert isfile(join(create_gtfs_project.project_base_path, "public_transport.sqlite")) is True



import unittest
import numpy as np
from aequilibrae.paths.AoN import copy_one_dimension, sum_axis1, linear_combination, linear_combination_skims
from aequilibrae.paths.AoN import copy_two_dimensions, copy_three_dimensions


class TestParallel(unittest.TestCase):
    def test_sum_axis1(self):
        target = np.zeros(50)
        source = np.random.rand(200).reshape(50, 4)
        b = np.sum(source, axis=1)

        sum_axis1(target, source, 1)
        self.assertEqual((b - target).max(), 0, "Sum Axis 1 failed")

    def test_linear_combination(self):
        target = np.zeros((50, 1))
        source = np.random.rand(50).reshape(50, 1)

        linear_combination(target, source, target, 0.05, 1)
        self.assertEqual((source - target * 20).max(), 0, "Linear combination failed")

    def test_linear_combination_skims(self):
        target = np.zeros((5, 5, 1))
        source = np.random.rand(25).reshape(5, 5, 1)

        linear_combination_skims(target, source, target, 0.8, 1)
        self.assertEqual((source - target * 5 / 4).max(), 0, "Linear combination failed")

    def test_triple_linear_combination(self):
        pass

    def test_triple_linear_combination_skims(self):
        pass

    def test_copy_one_dimension(self):
        target = np.zeros(50)
        source = np.random.rand(50)

        copy_one_dimension(target, source, 1)

        self.assertEqual(target.sum(), source.sum(), "Copying one dimension returned different values")
        if target.sum() == 0:
            self.fail("Target and source are the other way around for copying one dimension")

    def test_copy_two_dimensions(self):
        target = np.zeros(50).reshape(10, 5)
        source = np.random.rand(50).reshape(10, 5)

        copy_two_dimensions(target, source, 1)

        self.assertEqual(target.sum(), source.sum(), "Copying one dimension returned different values")
        if target.sum() == 0:
            self.fail("Target and source are the other way around for copying one dimension")

    def test_copy_three_dimensions(self):
        target = np.zeros(50).reshape(5, 2, 5)
        source = np.random.rand(50).reshape(5, 2, 5)

        copy_three_dimensions(target, source, 1)

        self.assertEqual(target.sum(), source.sum(), "Copying one dimension returned different values")
        if target.sum() == 0:
            self.fail("Target and source are the other way around for copying one dimension")


if __name__ == "__main__":
    unittest.main()



import os
import pytest

from aequilibrae.transit.lib_gtfs import GTFSRouteSystemBuilder


@pytest.fixture
def gtfs_file(create_path):
    return os.path.join(create_path, "gtfs_coquimbo.zip")


@pytest.fixture
def system_builder(transit_conn, gtfs_file):
    yield GTFSRouteSystemBuilder(
        network=transit_conn, agency_identifier="LISERCO, LISANCO, LINCOSUR", file_path=gtfs_file
    )


def test_set_capacities(system_builder):
    system_builder.set_capacities({0: [150, 300], 3: [42, 56]})
    assert system_builder.gtfs_data.__dict__["__capacities__"] == {0: [150, 300], 3: [42, 56]}


def test_dates_available(system_builder):
    dates = system_builder.dates_available()
    assert type(dates) == list


def test_set_allow_map_match(system_builder):
    assert system_builder.__dict__["_GTFSRouteSystemBuilder__do_execute_map_matching"] is False
    system_builder.set_allow_map_match(True)
    assert system_builder.__dict__["_GTFSRouteSystemBuilder__do_execute_map_matching"] is True


def test_map_match_tuple_exception(system_builder):
    with pytest.raises(TypeError):
        system_builder.map_match(route_types=3)


def test_map_match_int_exception(system_builder):
    with pytest.raises(TypeError):
        system_builder.map_match(route_types=[3.5])


def test_map_match(transit_conn, system_builder):
    system_builder.load_date("2016-04-13")
    system_builder.set_allow_map_match(True)
    system_builder.map_match()
    system_builder.save_to_disk()

    assert transit_conn.execute("SELECT * FROM pattern_mapping;").fetchone()[0] > 1


def test_set_agency_identifier(system_builder):
    assert system_builder.gtfs_data.agency.agency != "CTA"
    system_builder.set_agency_identifier("CTA")
    assert system_builder.gtfs_data.agency.agency == "CTA"


def test_set_feed(gtfs_file, system_builder):
    system_builder.set_feed(gtfs_file)
    assert system_builder.gtfs_data.archive_dir == gtfs_file


def test_set_description(system_builder):
    system_builder.set_description("CTA2019 fixed by John Doe after strong coffee")
    assert system_builder.description == "CTA2019 fixed by John Doe after strong coffee"


def test_set_date(system_builder):
    system_builder.set_date("2016-04-13")
    assert system_builder.__target_date__ == "2016-04-13"


def test_load_date(system_builder):
    system_builder.load_date("2016-04-13")
    assert system_builder.gtfs_data.agency.service_date == "2016-04-13"
    assert "101387" in system_builder.select_routes.keys()


def test_load_date_srid_exception(system_builder):
    system_builder.srid = None
    with pytest.raises(ValueError):
        system_builder.load_date("2016-04-13")


def test_load_date_not_available_date_exception(system_builder):
    with pytest.raises(ValueError):
        system_builder.load_date("2020-06-01")


def test_save_to_disk(transit_conn, system_builder):
    system_builder.load_date("2016-04-13")
    system_builder.save_to_disk()

    assert len(transit_conn.execute("SELECT * FROM route_links").fetchall()) == 78
    assert len(transit_conn.execute("SELECT * FROM trips;").fetchall()) == 360
    assert len(transit_conn.execute("SELECT * FROM routes;").fetchall()) == 2



import os
import uuid
from shutil import copytree
from tempfile import gettempdir
from unittest import TestCase

import numpy as np

from aequilibrae.distribution import GravityCalibration
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.parameters import Parameters
from aequilibrae.project import Project
from ...data import siouxfalls_project

zones = 100

# Impedance matrix_procedures
name_test = AequilibraeMatrix().random_name()

args = {"file_name": name_test, "zones": zones, "matrix_names": ["impedance"]}

impedance = AequilibraeMatrix()
impedance.create_empty(**args)
impedance.matrices[:, :, 0] = np.random.rand(zones, zones) * 1000
impedance.index[:] = np.arange(impedance.zones) + 100
impedance.computational_view(["impedance"])

args["matrix_names"] = ["base_matrix"]
args["memory_only"] = True
matrix = AequilibraeMatrix()
matrix.create_empty(**args)
matrix.matrices[:, :, 0] = np.random.rand(zones, zones) * 1000
matrix.index[:] = np.arange(matrix.zones) + 100
matrix.computational_view(["base_matrix"])


class TestGravityCalibration(TestCase):
    def test_calibrate(self):
        par = Parameters()

        args = {"impedance": impedance, "matrix": matrix, "function": "power", "nan_to_zero": False}

        distributed_matrix = GravityCalibration(par, **args)
        distributed_matrix.calibrate()
        if distributed_matrix.gap > 0.0001:
            self.fail("Calibration did not converge")

    def test_calibrate_with_omx(self):
        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)
        self.proj = Project()
        self.proj.open(self.proj_dir)
        self.par = self.proj.project_parameters

        mats = self.proj.matrices
        mats.update_database()

        imped = mats.get_matrix("omx")
        imped.computational_view(["free_flow_time"])

        mat = mats.get_matrix("SiouxFalls_omx")
        mat.computational_view("matrix")

        args = {"impedance": imped, "matrix": mat, "function": "power", "nan_to_zero": False}

        distributed_matrix = GravityCalibration(self.par, **args)
        distributed_matrix.calibrate()
        if distributed_matrix.gap > 0.0001:
            self.fail("Calibration did not converge")

        args = {"impedance": imped, "matrix": mat, "function": "power", "nan_to_zero": True}

        distributed_matrix = GravityCalibration(self.par, **args)
        distributed_matrix.calibrate()
        if distributed_matrix.gap > 0.0001:
            self.fail("Calibration did not converge")



import os
import sqlite3
import zipfile
from os.path import join

import pandas as pd
from pyproj import Transformer

# from aequilibrae.tools.geo import Geo
from aequilibrae.transit.functions.get_srid import get_srid
from aequilibrae.log import logger
from aequilibrae.transit.gtfs_writer import write_routes, write_agencies, write_fares
from aequilibrae.transit.gtfs_writer import write_stops, write_trips, write_stop_times, write_shapes
from aequilibrae.transit.route_system_reader import read_agencies, read_patterns
from aequilibrae.transit.route_system_reader import read_stop_times, read_stops, read_trips, read_routes
from aequilibrae.project.database_connection import database_connection


class RouteSystem:
    def __init__(self, database_path):
        self.__database_path = database_path
        self.__conn: sqlite3.Connection = None

        self.agencies = []
        self.stops = []
        self.routes = []
        self.trips = []
        self.patterns = []
        self.stop_times = pd.DataFrame([])

        self.fare_attributes = []
        self.fare_rules = []
        self.zones = []

        self.transformer = Transformer.from_crs(f"epsg:{get_srid()}", "epsg:4326", always_xy=True)

    def load_route_system(self):
        self._read_agencies()
        self._read_stops()
        self._read_routes()
        self._read_patterns()
        self._read_trips()
        self._read_stop_times()

    def _read_agencies(self):
        self.agencies = read_agencies(self.conn)

    def _read_stops(self):
        self.stops = read_stops(self.conn, self.transformer)

    def _read_routes(self):
        self.routes = read_routes(self.conn)

    def _read_patterns(self):
        self.patterns = self.patterns or read_patterns(self.conn, self.transformer)

    def _read_trips(self):
        self.trips = self.trips or read_trips(self.conn)

    def _read_stop_times(self):
        self.stop_times = read_stop_times(self.conn)

    @property
    def conn(self) -> sqlite3.Connection:
        self.__conn = self.__conn or database_connection(join(self.__database_path, "public_transport.sqlite"))
        return self.__conn

    def write_GTFS(self, path_to_folder: str):
        """ """
        # timezone = self._timezone()

        write_agencies(self.agencies, path_to_folder)
        write_stops(self.stops, path_to_folder)
        write_routes(self.routes, path_to_folder)
        write_shapes(self.patterns, path_to_folder)

        write_trips(self.trips, path_to_folder, self.conn)
        write_stop_times(self.stop_times, path_to_folder)
        write_fares(path_to_folder, self.conn)
        self._zip_feed(path_to_folder)

    # def _timezone(self, allow_error=True):
    #     geotool = Geo()
    #     geotool.conn = self.conn
    #     try:
    #         return geotool.get_timezone()
    #     except Exception as e:
    #         logger.error("Could not retrieve the correct time zone for GTFS exporter. Using Chicago instead")
    #         if not allow_error:
    #             raise e
    #     return "America/Chicago"

    def _zip_feed(self, path_to_folder: str):
        filename = join(path_to_folder, "polaris_gtfs.zip")
        files = [
            "agency",
            "stops",
            "routes",
            "trips",
            "stop_times",
            "calendar",
            "shapes",
            "fare_attributes",
            "fare_rules",
        ]
        with zipfile.ZipFile(filename, mode="w", compression=zipfile.ZIP_DEFLATED) as zip_file:
            for file in files:
                zip_file.write(join(path_to_folder, f"{file}.txt"), f"{file}.txt")
                os.unlink(join(path_to_folder, f"{file}.txt"))



from os.path import join, dirname, abspath
import pytest

import pandas as pd
from aequilibrae.transit import Transit

from aequilibrae.transit.gtfs_loader import GTFSReader
from aequilibrae.utils.create_example import create_example


@pytest.fixture
def gtfs_loader(create_gtfs_project):
    return GTFSReader()


@pytest.fixture
def gtfs_fldr(create_path):
    return join(create_path, "gtfs_coquimbo.zip")


def test_set_feed_path(gtfs_loader, gtfs_fldr):
    gtfs = gtfs_loader

    with pytest.raises(Exception):
        gtfs.set_feed_path(gtfs_fldr + "_")


def test_load_data(gtfs_loader, gtfs_fldr):
    cap = pd.read_csv(join(abspath(dirname("tests")), "tests/data/gtfs/transit_max_speeds.txt"))

    df = cap[cap.city == "Coquimbo"]
    df.loc[df.min_distance < 100, "speed"] = 10
    dict_speeds = {x: df for x, df in df.groupby(["mode"])}
    gtfs = gtfs_loader

    gtfs._set_maximum_speeds(dict_speeds)
    gtfs.set_feed_path(gtfs_fldr)
    gtfs.load_data("2016-04-13")



import os
import uuid
from copy import copy, deepcopy
from random import randint
from shutil import copytree, rmtree
from tempfile import gettempdir
from unittest import TestCase

from aequilibrae.project import Project
from ...data import siouxfalls_project


class TestLinks(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)
        self.network = self.project.network
        self.curr = self.project.conn.cursor()

    def tearDown(self) -> None:
        self.project.close()
        del self.curr
        try:
            rmtree(self.proj_dir)
        except Exception as e:
            print(f"Failed to remove at {e.args}")

    def test_get(self):
        links = self.network.links
        with self.assertRaises(ValueError):
            _ = links.get(123456)

        link = links.get(1)
        self.assertEqual(link.capacity_ab, 25900.20064, "Did not populate link correctly")

    def test_new(self):
        links = self.network.links
        new_link = links.new()

        self.curr.execute("Select max(link_id) + 1 from Links")
        id = self.curr.fetchone()[0]
        self.assertEqual(new_link.link_id, id, "Did not populate new link ID properly")
        self.assertEqual(new_link.geometry, None, "Did not populate new geometry properly")

    def test_copy_link(self):
        links = self.network.links

        with self.assertRaises(ValueError):
            _ = links.copy_link(11111)

        new_link = links.copy_link(11)

        old_link = links.get(11)

        self.assertEqual(new_link.geometry, old_link.geometry)
        self.assertEqual(new_link.a_node, old_link.a_node)
        self.assertEqual(new_link.b_node, old_link.b_node)
        self.assertEqual(new_link.direction, old_link.direction)
        self.assertEqual(new_link.distance, old_link.distance)
        self.assertEqual(new_link.modes, old_link.modes)
        self.assertEqual(new_link.link_type, old_link.link_type)
        new_link.save()

    def test_delete(self):
        links = self.network.links

        _ = links.get(10)

        self.curr.execute("Select count(*) from Links")
        tot = self.curr.fetchone()[0]
        links.delete(10)
        links.delete(11)
        self.curr.execute("Select count(*) from Links")
        tot2 = self.curr.fetchone()[0]
        self.assertEqual(tot, tot2 + 2, "Did not delete the link properly")

        with self.assertRaises(ValueError):
            links.delete(123456)

        with self.assertRaises(ValueError):
            _ = links.get(10)

    def test_fields(self):
        links = self.network.links
        f_editor = links.fields

        fields = sorted(f_editor.all_fields())
        self.curr.execute("pragma table_info(links)")
        dt = self.curr.fetchall()

        actual_fields = set([x[1].replace("_ab", "").replace("_ba", "") for x in dt if x[1] != "ogc_fid"])
        actual_fields = sorted(list(actual_fields))

        self.assertEqual(fields, actual_fields, "Table editor is weird for table links")

    def test_refresh(self):
        links = self.network.links

        link1 = links.get(1)
        val = randint(1, 99999999)
        original_value = link1.capacity_ba

        link1.capacity_ba = val
        link1_again = links.get(1)
        self.assertEqual(link1_again.capacity_ba, val, "Did not preserve correctly")

        links.refresh()
        link1 = links.get(1)
        self.assertEqual(link1.capacity_ba, original_value, "Did not reset correctly")

    def test_copy(self):
        nodes = self.network.nodes
        with self.assertRaises(Exception):
            _ = copy(nodes)
        with self.assertRaises(Exception):
            _ = deepcopy(nodes)



from random import randint, choice
import pytest

from aequilibrae.transit.transit_elements import Trip
from tests.aequilibrae.transit.random_word import randomword


class TestTrip:
    @pytest.fixture
    def data(self):
        return {
            "route_id": randomword(randint(0, 40)),
            "service_id": randomword(randint(0, 40)),
            "trip_id": randomword(randint(0, 40)),
            "trip_headsign": randomword(randint(0, 5)),
            "trip_short_name": randomword(randint(0, 5)),
            "block_id": randomword(randint(0, 5)),
            "shape_id": randomword(randint(0, 5)),
            "direction_id": choice([0, 1]),
            "bikes_allowed": choice([0, 1]),
        }

    def test_populate(self, data):
        s = Trip()

        s._populate(tuple(data.values()), list(data.keys()))
        data["route"] = data.pop("route_id")
        data["trip"] = data.pop("trip_id")
        for key, val in s.__dict__.items():
            if key in data:
                assert val == data[key], "StopTime population with record failed"

        data[randomword(randint(1, 30))] = randomword(randint(1, 30))

        with pytest.raises(KeyError):
            s = Trip()
            s._populate(tuple(data.values()), list(data.keys()))

    def test_save_to_database(self, data, transit_conn):
        r = Trip()
        r._populate(tuple(data.values()), list(data.keys()))
        times = [r for r in range(randint(5, 15))]
        patid = randint(15, 2500000)

        r.arrivals = [r for r in times]
        r.departures = [r for r in times]
        r.pattern_id = patid
        r.source_time = [0] * len(times)
        r.save_to_database(transit_conn)

        result = transit_conn.execute("Select pattern_id from trips where trip_id=?", [r.trip_id]).fetchone()[0]
        assert result == patid, "Saving trip to trips failed"

        records, counter = transit_conn.execute(
            'Select count(*), max("seq") from trips_schedule where trip_id=?', [r.trip_id]
        ).fetchone()
        assert records == len(times), "Saving trip to trips_schedule failed"
        assert counter == max(times), "Saving trip to trips_schedule failed"



from math import pi
import sqlite3
from random import randint
import pytest

from shapely.geometry import LineString
from aequilibrae.transit.functions.get_srid import get_srid

from aequilibrae.transit.transit_elements import Link
from tests.aequilibrae.transit.random_word import randomword


class TestLink:
    @pytest.fixture
    def srid(self):
        return get_srid()

    def test_build_object(self, srid):
        new_link = Link(srid)

        assert new_link.key == "####-1", "Pair not initiated properly"
        assert new_link.length == -1, "Length not initiated properly"

        assert new_link.srid == srid, "SRID was not assigned properly"

        fstop = randomword(randint(3, 15))
        tstop = randomword(randint(3, 15))

        new_link.from_stop = fstop
        assert new_link.key == fstop + "####-1", "Pair not computed properly"

        new_link.to_stop = tstop
        assert new_link.key == fstop + "##" + tstop + "##-1", "Pair not computed properly"

        geo = LineString([(0, 0), (3, 4)])

        new_link.geo = geo
        assert new_link.length == pytest.approx(5 * pi * 6371000 / 180), "Length not computed properly"

    def test_save_to_database(self, srid, transit_conn):
        route_type = randint(0, 13)
        fstop = randomword(randint(3, 15))
        tstop = randomword(randint(3, 15))
        geo = LineString([(0, 0), (3, 4)])

        new_link = Link(srid)

        with pytest.raises(AttributeError):
            new_link.save_to_database(transit_conn)

        new_link.geo = geo
        new_link.transit_link = 10000001
        new_link.type = route_type
        new_link.from_stop = fstop
        new_link.to_stop = tstop

        with pytest.raises(sqlite3.IntegrityError):
            new_link.save_to_database(transit_conn)

        new_link.pattern_id = 10001001000
        new_link.seq = 4

        new_link.save_to_database(transit_conn)

        from_stop, to_stop, dist = transit_conn.execute(
            "Select from_stop, to_stop, distance from route_links where from_stop=?", [fstop]
        ).fetchone()
        assert [from_stop, to_stop, round(dist * 180 / pi / 6371000)] == [
            fstop,
            tstop,
            geo.length,
        ], "Saving link to the database failed"



from unittest import TestCase
from shapely.ops import substring
from copy import copy, deepcopy
import os
from shutil import copytree, rmtree
import uuid
from random import randint, random
from tempfile import gettempdir
from aequilibrae.project import Project

from ...data import siouxfalls_project


class TestLink(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "../../temp_data") + ";" + os.environ["PATH"]

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)
        self.network = self.project.network
        self.curr = self.project.conn.cursor()

        self.links = self.network.links
        self.modes = self.network.modes
        self.lid = randint(1, 24)
        self.link = self.links.get(self.lid)

    def tearDown(self) -> None:
        self.project.close()
        del self.curr
        try:
            rmtree(self.proj_dir)
        except Exception as e:
            print(f"Failed to remove at {e.args}")

    def test_delete(self):
        self.link.delete()

        with self.assertRaises(Exception):
            _ = self.links.get(self.lid)

        self.curr.execute("Select count(*) from links where link_id=?", [self.lid])

        self.assertEqual(0, self.curr.fetchone()[0], f"Failed to delete link {self.lid}")

    def test_save(self):
        self.link.save()
        extension = random()
        name = "just a non-important value"

        geo = substring(self.link.geometry, 0, extension, normalized=True)

        self.link.name = name
        self.link.geometry = geo

        self.link.save()
        self.links.refresh()
        link2 = self.links.get(self.lid)

        self.assertEqual(link2.name, name, "Failed to save the link name")
        self.assertAlmostEqual(link2.geometry, geo, 3, "Failed to save the link geometry")

        tot_prev = self.network.count_links()
        lnk = self.links.new()
        lnk.geometry = substring(self.link.geometry, 0, 0.88, normalized=True)
        lnk.modes = "c"
        lnk.save()

        self.assertEqual(tot_prev + 1, self.network.count_links(), "Failed to save new link")

    def test_set_modes(self):
        self.link.set_modes("cbt")

        self.assertEqual(self.link.modes, "cbt", "Did not set modes correctly")
        self.link.save()

        self.assertEqual(self.__check_mode(), "cbt")

    def test_add_mode(self):
        for mode in [1, ["cbt"]]:
            with self.assertRaises(TypeError):
                self.link.add_mode(mode)
        with self.assertRaises(ValueError):
            self.link.add_mode("bt")

        self.link.add_mode("b")
        self.link.save()
        self.assertEqual(self.__check_mode(), "cb")

        mode = self.modes.get("t")
        self.link.add_mode(mode)
        self.link.save()
        self.assertEqual(self.__check_mode(), "cbt")

    def test_drop_mode(self):
        self.link.set_modes("cbt")
        self.link.save()
        self.assertEqual(self.__check_mode(), "cbt")

        self.link.drop_mode("t")
        self.link.save()
        self.assertEqual(self.__check_mode(), "cb")

        mode = self.modes.get("b")
        self.link.drop_mode(mode)
        self.link.save()
        self.assertEqual(self.__check_mode(), "c")

    def test_data_fields(self):
        link2 = self.links.get(randint(1, 24))
        while link2.link_id == self.link.link_id:
            link2 = self.links.get(randint(1, 24))

        self.assertEqual(link2.data_fields(), self.link.data_fields(), "Different links have different data fields")

        fields = sorted(link2.data_fields())
        self.curr.execute("pragma table_info(links)")
        dt = self.curr.fetchall()

        data_fields = sorted([x[1] for x in dt if x[1] != "ogc_fid"])

        self.assertEqual(sorted(fields), sorted(data_fields), "Link has unexpected set of fields")

    def __check_mode(self):
        sql = "Select modes from links where link_id=?"
        self.curr.execute(sql, [self.lid])
        return self.curr.fetchone()[0]



from unittest import TestCase
import tempfile
import os
from random import choice
from warnings import warn
from shutil import copytree, rmtree
from aequilibrae.project import Project
import uuid
import sqlite3
from aequilibrae import logger
from ...data import no_triggers_project


class TestProject(TestCase):
    def setUp(self) -> None:
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(no_triggers_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

        # Modes to add
        sql = "INSERT INTO modes (mode_name, mode_id) VALUES (?, ?);"
        for mid in ["p", "l", "g", "x", "y", "d", "k", "a", "r", "n", "m"]:
            self.curr.execute(sql, [f"mode_{mid}", mid])

        self.proj.conn.commit()

        root = os.path.dirname(os.path.realpath(__file__)).replace("tests", "")
        qry_file = os.path.join(root, "database_specification/network/triggers/link_type_table_triggers.sql")
        with open(qry_file, "r") as sql_file:
            self.queries = sql_file.read()
        self.queries = [cmd for cmd in self.queries.split("#")]

        curr = self.proj.conn.cursor()
        self.rtree = True
        try:
            curr.execute("SELECT rtreecheck('idx_nodes_geometry');")
        except Exception as e:
            self.rtree = False
            warn(f"RTREE not available --> {e.args}")

    def tearDown(self) -> None:
        self.proj.close()

    def __get_query(self, qry: str) -> str:
        for query in self.queries:
            if qry in query:
                return query
        raise FileNotFoundError("QUERY DOES NOT EXIST")

    def test_all_tests_considered(self):
        tests_added = list(self.__dir__())
        tests_added = [x[5:] for x in tests_added if x[:5] == "test_"]

        for trigger in self.queries:
            if "TRIGGER" in trigger.upper():
                found = [x for x in tests_added if x in trigger]
                if not found:
                    self.fail(f"Trigger not tested. {trigger}")

    def test_link_type_single_letter_update(self):
        cmd = self.__get_query("link_type_single_letter_update")
        sql = "UPDATE 'link_types' SET link_type_id= 'ttt' where link_type_id='t'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "UPDATE 'link_types' SET link_type_id= 'ww' where link_type_id='w'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_link_type_single_letter_insert(self):
        cmd = self.__get_query("link_type_single_letter_insert")
        sql = "INSERT INTO 'link_types' (link_type, link_type_id) VALUES(?, ?)"
        self.curr.execute(sql, ["test1a", "more_than_one"])

        self.curr.execute(cmd)

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql, ["test1b", "mm"])

    def test_link_type_keep_if_in_use_updating(self):
        cmd = self.__get_query("link_type_keep_if_in_use_updating")

        sql = "UPDATE 'link_types' SET link_type= 'ttt' where link_type='test'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "UPDATE 'link_types' SET link_type= 'QQQ' where link_type='test2'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_link_type_keep_if_in_use_deleting(self):
        cmd = self.__get_query("link_type_keep_if_in_use_deleting")

        sql = "DELETE FROM 'link_types' where link_type='test3'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "DELETE FROM 'link_types' where link_type='test4'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_link_type_on_links_update(self):
        cmd = self.__get_query("link_type_on_links_update")

        sql = "UPDATE 'links' SET link_type= 'rrr' where link_type='test3'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "UPDATE 'links' SET link_type= 'not_valid_type' where link_type='test4'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_link_type_on_links_insert(self):
        cmd = self.__get_query("link_type_on_links_insert")

        if self.rtree:
            self.curr.execute("pragma table_info(links)")
            f = self.curr.fetchall()
            fields = {x[1]: x[0] for x in f}

            sql = "select * from links where link_id=70"
            self.curr.execute(sql)
            a = [x for x in self.curr.fetchone()]
            a[fields["link_type"]] = "something indeed silly123"
            a[fields["link_id"]] = 456789
            a[fields["a_node"]] = 777
            a[fields["b_node"]] = 999
            a[0] = 456789

            idx = ",".join(["?"] * len(a))
            self.curr.execute(f"insert into links values ({idx})", a)
            self.curr.execute("delete from links where link_id=456789")

            self.curr.execute(cmd)

            with self.assertRaises(sqlite3.IntegrityError):
                self.curr.execute(f"insert into links values ({idx})", a)

            self.curr.execute("select link_type from link_types;")
            a[fields["link_type"]] = self.curr.fetchone()[0]
            self.curr.execute(f"insert into links values ({idx})", a)

    def test_link_type_on_links_delete_protected_link_type(self):
        cmd = self.__get_query("link_type_on_links_delete_protected_link_type")

        self.curr.execute(cmd)
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute('delete from link_types where link_type_id="z"')

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute('delete from link_types where link_type_id="y"')

    def test_link_type_id_keep_if_protected_type(self):
        cmd = self.__get_query("link_type_id_keep_if_protected_type")

        self.curr.execute(cmd)

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute('update link_types set link_type_id="x" where link_type_id="y"')

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute('update link_types set link_type_id="x" where link_type_id="z"')

    def test_link_type_keep_if_protected_type(self):
        cmd = self.__get_query("link_type_keep_if_protected_type")
        self.curr.execute(cmd)

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute('update link_types set link_type="xsdfg" where link_type_id="z"')

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute('update link_types set link_type="xsdfg" where link_type_id="y"')

    def test_link_type_on_nodes_table_update_nodes_link_type(self):
        cmd = self.__get_query("link_type_on_nodes_table_update_nodes_link_type")
        self.curr.execute(cmd)

        self.curr.execute('update nodes set link_types="qwerrreyrtuyiuio" where node_id=1')

        self.curr.execute("select link_types from nodes where node_id=1")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "etuw", "link_types was allowed to be corrupted in the nodes table")

    def test_link_type_on_nodes_table_update_links_link_type(self):
        cmd = self.__get_query("link_type_on_nodes_table_update_links_link_type")
        self.curr.execute(cmd)

        self.curr.execute('update links set link_type="test" where link_id=15')

        self.curr.execute("select link_types from nodes where node_id=6")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "rtw", "link_types on nodes table not updated with new link type in the links")

        self.curr.execute("select link_types from nodes where node_id=5")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "ertw", "link_types was allowed to be corrupted in the nodes table")

    def test_link_type_on_nodes_table_update_links_a_node(self):
        cmd = self.__get_query("link_type_on_nodes_table_update_links_a_node")
        self.curr.execute(cmd)

        self.curr.execute("update links set a_node=1 where link_id=15")

        self.curr.execute("select link_types from nodes where node_id=1")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "etuw", "link_types on nodes table not updated with new link type in the links")

        self.curr.execute("select link_types from nodes where node_id=6")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "rw", "link_types was allowed to be corrupted in the nodes table")

    def test_link_type_on_nodes_table_update_links_b_node(self):
        cmd = self.__get_query("link_type_on_nodes_table_update_links_b_node")
        self.curr.execute(cmd)

        self.curr.execute("update links set b_node=1 where link_id=15")

        self.curr.execute("select link_types from nodes where node_id=1")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "etuw", "link_types on nodes table not updated with new link type in the links")

        self.curr.execute("select link_types from nodes where node_id=5")
        lts = self.curr.fetchone()[0]

        self.assertEqual(lts, "ertw", "link_types was allowed to be corrupted in the nodes table")



import os
import shutil

from aequilibrae.log import logger

from aequilibrae.project.project_creation import initialize_tables
from aequilibrae.reference_files import spatialite_database
from aequilibrae.transit.lib_gtfs import GTFSRouteSystemBuilder


class Transit:
    default_capacities = {
        0: [150, 300],  # Tram, Streetcar, Light rail
        1: [280, 560],  # Subway/metro
        2: [700, 700],  # Rail
        3: [30, 60],  # Bus
        4: [400, 800],  # Ferry
        5: [20, 40],  # Cable tram
        11: [30, 60],  # Trolleybus
        12: [50, 100],  # Monorail
        "other": [30, 60],
    }

    def __init__(self, project):
        """
        :Arguments:
             **project** (:obj:`Project`, optional): The Project to connect to. By default, uses the currently
            active project
        """

        self.project_base_path = project.project_base_path
        self.logger = logger
        self.__transit_file = os.path.join(project.project_base_path, "public_transport.sqlite")

        self.create_transit_database()

    def new_gtfs_builder(self, agency, file_path, day="", description="") -> GTFSRouteSystemBuilder:
        """Returns a GTFSRouteSystemBuilder object compatible with the project

        :Arguments:
            **agency** (:obj:`str`): Name for the agency this feed refers to (e.g. 'CTA')

            **file_path** (:obj:`str`): Full path to the GTFS feed (e.g. 'D:/project/my_gtfs_feed.zip')

            **day** (:obj:`str`, *Optional*): Service data contained in this field to be imported (e.g. '2019-10-04')

            **description** (:obj:`str`, *Optional*): Description for this feed (e.g. 'CTA2019 fixed by John Doe')

        :Return:
            **gtfs_feed** (:obj:`StaticGTFS`): A GTFS feed that can be added to this network
        """
        gtfs = GTFSRouteSystemBuilder(
            network=self.project_base_path,
            agency_identifier=agency,
            file_path=file_path,
            day=day,
            description=description,
            default_capacities=self.default_capacities,
        )
        return gtfs

    def create_transit_database(self):
        """Creates the public transport database"""
        if not os.path.exists(self.__transit_file):
            shutil.copyfile(spatialite_database, self.__transit_file)
            initialize_tables(self, "transit")



from unittest import TestCase
from aequilibrae.distribution import SyntheticGravityModel
import tempfile
import os

filename = os.path.join(tempfile.gettempdir(), "aequilibrae_model_test.mod")


class TestSyntheticGravityModel(TestCase):
    def test_load(self):
        self.test_save()
        model = SyntheticGravityModel()
        model.load(filename)

        if model.alpha != 0.1:
            self.fail("Gravity model: Alpha not saved properly")

        if model.function != "POWER":
            self.fail("Gravity model: Function not saved properly")

        if model.beta is not None:
            self.fail("Gravity model: Beta not saved properly")

    def test_save(self):
        model = SyntheticGravityModel()
        model.function = "EXPO"
        model.beta = 0.1
        self.assertEqual(model.function, "EXPO")  # Did we save the value?

        model.function = "POWER"
        # Check if we zeroed the parameters when changing the function
        self.assertEqual(model.beta, None)
        model.alpha = 0.1

        model.save(filename)



import os
import random
import sqlite3
import string
import uuid
from random import choice
from tempfile import gettempdir
from unittest import TestCase

import numpy as np
import pandas as pd

from aequilibrae import TrafficAssignment, TrafficClass, Graph
from aequilibrae.utils.create_example import create_example
from ...data import siouxfalls_project


class TestTrafficAssignment(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        proj_path = os.path.join(gettempdir(), "test_traffic_assignment_" + uuid.uuid4().hex)
        self.project = create_example(proj_path)
        self.project.network.build_graphs()
        self.car_graph = self.project.network.graphs["c"]  # type: Graph
        self.car_graph.set_graph("free_flow_time")
        self.car_graph.set_blocked_centroid_flows(False)

        self.matrix = self.project.matrices.get_matrix("demand_omx")
        self.matrix.computational_view()

        self.assignment = TrafficAssignment()
        self.assigclass = TrafficClass("car", self.car_graph, self.matrix)

        self.algorithms = ["msa", "cfw", "bfw", "frank-wolfe"]

    def test_matrix_with_wrong_type(self):
        self.matrix.matrix_view = np.array(self.matrix.matrix_view, np.int32)
        with self.assertRaises(TypeError):
            _ = TrafficClass("car", self.car_graph, self.matrix)

    def test_set_vdf(self):
        with self.assertRaises(ValueError):
            self.assignment.set_vdf("CQS")

        self.assignment.set_vdf("BPR")

    def test_set_classes(self):
        with self.assertRaises(AttributeError):
            self.assignment.set_classes([1, 2])

        with self.assertRaises(Exception):
            self.assignment.set_classes(self.assigclass)

        self.assignment.set_classes([self.assigclass])
        # self.fail()

    def test_algorithms_available(self):
        algs = self.assignment.algorithms_available()
        real = ["all-or-nothing", "msa", "frank-wolfe", "bfw", "cfw", "fw"]

        diff = [x for x in real if x not in algs]
        diff2 = [x for x in algs if x not in real]

        if len(diff) + len(diff2) > 0:
            self.fail("list of algorithms raised is wrong")

    def test_set_cores(self):
        with self.assertRaises(Exception):
            self.assignment.set_cores(3)

        self.assignment.add_class(self.assigclass)
        with self.assertRaises(ValueError):
            self.assignment.set_cores("q")

        self.assignment.set_cores(3)

    def test_set_algorithm(self):
        with self.assertRaises(AttributeError):
            self.assignment.set_algorithm("not an algo")

        self.assignment.add_class(self.assigclass)

        with self.assertRaises(Exception):
            self.assignment.set_algorithm("msa")

        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 10

        for algo in self.algorithms:
            for _ in range(10):
                algo = "".join([x.upper() if random.random() < 0.5 else x.lower() for x in algo])
                self.assignment.set_algorithm(algo)

        with self.assertRaises(AttributeError):
            self.assignment.set_algorithm("not a valid algorithm")

    def test_set_vdf_parameters(self):
        with self.assertRaises(Exception):
            self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_vdf("bpr")
        self.assignment.add_class(self.assigclass)
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

    def test_set_time_field(self):
        with self.assertRaises(ValueError):
            self.assignment.set_time_field("capacity")

        self.assignment.add_class(self.assigclass)

        N = random.randint(1, 50)
        val = "".join(random.choices(string.ascii_uppercase + string.digits, k=N))
        with self.assertRaises(ValueError):
            self.assignment.set_time_field(val)

        self.assignment.set_time_field("free_flow_time")
        self.assertEqual(self.assignment.time_field, "free_flow_time")

    def test_set_capacity_field(self):
        with self.assertRaises(ValueError):
            self.assignment.set_capacity_field("capacity")

        self.assignment.add_class(self.assigclass)

        N = random.randint(1, 50)
        val = "".join(random.choices(string.ascii_uppercase + string.digits, k=N))
        with self.assertRaises(ValueError):
            self.assignment.set_capacity_field(val)

        self.assignment.set_capacity_field("capacity")
        self.assertEqual(self.assignment.capacity_field, "capacity")

    def test_execute_and_save_results(self):
        conn = sqlite3.connect(os.path.join(siouxfalls_project, "project_database.sqlite"))
        results = pd.read_sql("select volume from links order by link_id", conn)

        self.assignment.add_class(self.assigclass)
        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 10
        self.assignment.set_algorithm("msa")
        self.assignment.execute()

        msa10_rgap = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume.values)[0, 1]
        self.assertLess(0.8, correl)

        self.assignment.max_iter = 500
        self.assignment.rgap_target = 0.001
        self.assignment.set_algorithm("msa")
        self.assignment.execute()
        msa25_rgap = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.98, correl)

        self.assignment.set_algorithm("frank-wolfe")
        self.assignment.execute()

        fw25_rgap = self.assignment.assignment.rgap
        fw25_iters = self.assignment.assignment.iter

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.99, correl)

        self.assignment.set_algorithm("cfw")
        self.assignment.execute()
        cfw25_rgap = self.assignment.assignment.rgap
        cfw25_iters = self.assignment.assignment.iter

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.995, correl)

        # For the last algorithm, we set skimming
        self.car_graph.set_skimming(["free_flow_time", "distance"])
        assigclass = TrafficClass("car", self.car_graph, self.matrix)
        self.assignment.set_classes([assigclass])

        self.assignment.set_algorithm("bfw")
        self.assignment.execute()
        bfw25_rgap = self.assignment.assignment.rgap
        bfw25_iters = self.assignment.assignment.iter

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.999, correl)

        self.assertLess(msa25_rgap, msa10_rgap)
        # MSA and FW do not reach 1e-4 within 500 iterations, cfw and bfw do
        self.assertLess(fw25_rgap, msa25_rgap)
        self.assertLess(cfw25_rgap, self.assignment.rgap_target)
        self.assertLess(bfw25_rgap, self.assignment.rgap_target)
        # we expect bfw to converge quicker than cfw
        self.assertLess(cfw25_iters, fw25_iters)
        self.assertLess(bfw25_iters, cfw25_iters)

        self.assignment.save_results("save_to_database")
        self.assignment.save_skims(matrix_name="all_skims", which_ones="all")

        with self.assertRaises(ValueError):
            self.assignment.save_results("save_to_database")

    def test_execute_no_project(self):
        conn = sqlite3.connect(os.path.join(siouxfalls_project, "project_database.sqlite"))
        results = pd.read_sql("select volume from links order by link_id", conn)

        self.project.close()

        assignment = TrafficAssignment()
        assignment.add_class(self.assigclass)
        assignment.set_vdf("BPR")
        assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        assignment.set_capacity_field("capacity")
        assignment.set_time_field("free_flow_time")

        assignment.max_iter = 10
        assignment.set_algorithm("msa")
        assignment.execute()

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume.values)[0, 1]
        self.assertLess(0.8, correl)

        with self.assertRaises(FileNotFoundError):
            assignment.save_results("anything")

    def test_info(self):
        iterations = random.randint(1, 10000)
        rgap = random.random() / 10000
        algo = choice(self.algorithms)

        self.assignment.add_class(self.assigclass)
        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = iterations
        self.assignment.rgap_target = rgap
        self.assignment.set_algorithm(algo)

        # TY
        for _ in range(10):
            algo = "".join([x.upper() if random.random() < 0.5 else x.lower() for x in algo])

        dct = self.assignment.info()
        if algo.lower() == "fw":
            algo = "frank-wolfe"
        self.assertEqual(dct["Algorithm"], algo.lower(), "Algorithm not correct in info method")

        self.assertEqual(dct["Maximum iterations"], iterations, "maximum iterations not correct in info method")



import pytest
from random import randint
from datetime import datetime, timedelta
from aequilibrae.transit.transit_elements import Service
from tests.aequilibrae.transit.random_word import randomword


class TestService:
    @pytest.fixture
    def today(self):
        tdy = datetime.today()
        today = tdy
        return today

    @pytest.fixture
    def past(self, today):
        past = today - timedelta(days=randint(1, 100))
        past = past
        return past

    @pytest.fixture
    def data_dict(self, past, today):
        return {
            "service_id": randomword(randint(0, 40)),
            "monday": 1,
            "tuesday": 1,
            "wednesday": 1,
            "thursday": 1,
            "friday": 1,
            "saturday": 1,
            "sunday": 1,
            "start_date": past,
            "end_date": today,
        }

    def test__populate(self, data_dict, today, past):
        s = Service()

        s._populate(tuple(data_dict.values()), list(data_dict.keys()))
        for key, val in s.__dict__.items():
            if key in data_dict:
                assert val == data_dict[key], "Service population with record failed"

        time_span = (today - past).days + 1  # n + 1 intervals

        assert time_span == len(s.dates), "Returned the wrong dates for service"

        # Test with no weekdays available
        for key, val in data_dict.items():
            if val == 1:
                data_dict[key] = 0

        s = Service()
        s._populate(tuple(data_dict.values()), list(data_dict.keys()))
        assert 0 == len(s.dates), "Returned too many dates for service"

        data_dict[randomword(randint(1, 15))] = randomword(randint(1, 20))
        s = Service()
        with pytest.raises(KeyError):
            s._populate(tuple(data_dict.values()), list(data_dict.keys()))



import math
from sqlite3 import Connection
from typing import List, Tuple, Optional

import numpy as np
import pandas as pd
import shapely.wkb
from shapely.geometry import LineString, Polygon
from shapely.ops import substring

from aequilibrae.log import logger
from aequilibrae.transit.functions.get_srid import get_srid
from aequilibrae.utils.geo_index import GeoIndex
from .basic_element import BasicPTElement
from .link import Link
from .mode_correspondence import mode_correspondence

DEAD_END_RUN = 40


class Pattern(BasicPTElement):
    """
    Represents a stop pattern for a particular route, as defined in GTFS.
    """

    def __init__(self, route_id, gtfs_feed) -> None:
        """
        Args:
            *pattern_id* (:obj:`str`): Pre-computed ID for this pattern
            *geotool* (:obj:`Geo`): Suite of geographic utilities.
        """
        self.pattern_hash = ""
        self.pattern_id = -1
        self.route_id = route_id
        self.route = ""
        self.agency_id = None
        self.longname = ""
        self.shortname = ""
        self.description = ""
        self.seated_capacity = None
        self.total_capacity = None
        self.__srid = get_srid()
        self.__geotool = gtfs_feed.geotool
        self.__geolinks = self.__geotool.network.links.data
        self.__logger = logger

        self.__feed = gtfs_feed
        # For map matching
        self.raw_shape: LineString = None
        self._stop_based_shape: LineString = None
        self.shape: LineString = None
        self.route_type: int = None
        self.links: List[Link] = []
        self.network_candidates = []
        self.full_path: List[int] = []
        self.fpath_dir: List[int] = []
        self.pattern_mapping = []
        self.stops = []
        self.__map_matching_error = {}

        self.__graph = None
        self.__res = None
        self.__curr_net_nodes_from_stops = []
        self.__net_links_from_stops = []
        self.__net_nodes_from_stops = []
        self.__mm_fail_position = -1
        self.__map_matched = False
        self.__match_quality = None
        self.shape_length = -1

    def save_to_database(self, conn: Connection, commit=True) -> None:
        """Saves the pattern to the routes table"""

        shp = self.best_shape()
        geo = None if shp is None else shp.wkb

        data = [
            self.pattern_id,
            self.route_id,
            self.route,
            self.agency_id,
            self.shortname,
            self.longname,
            self.description,
            self.route_type,
            self.seated_capacity,
            self.total_capacity,
            geo,
            self.__srid,
        ]

        sql = """insert into routes (pattern_id, route_id, route, agency_id, shortname, longname, description, route_type, seated_capacity,
                         total_capacity, geometry) values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ST_Multi(GeomFromWKB(?, ?)));"""
        conn.execute(sql, data)

        if self.pattern_mapping and self.shape:
            sqlgeo = """insert into pattern_mapping (pattern_id, seq, link, dir, geometry)
                        values (?, ?, ?, ?, GeomFromWKB(?, ?));"""
            sql = """insert into pattern_mapping (pattern_id, seq, link, dir)
                                                  values (?, ?, ?, ?);"""

            for record in self.pattern_mapping:
                if record[-1] is None:
                    conn.execute(sql, record[:-1])
                else:
                    geo = shapely.wkb.loads(record[-1])
                    if isinstance(geo, LineString):
                        conn.execute(sqlgeo, record + [self.__srid])
                    else:
                        conn.execute(sql, record[:-1])

    def best_shape(self) -> LineString:
        """Gets the best version of shape available for this pattern"""
        shp = self._stop_based_shape if self.raw_shape is None else self.raw_shape
        shp = shp if self.shape is None else self.shape
        return shp

    def map_match(self):
        """Map matches the route into the network, considering its appropriate shape.

        Part of the map-matching process is to find the network links corresponding the pattern's
        raw shape, so that method will be called in case it has not been called before.

        The basic algorithm behind the map-matching algorithm is described in https://doi.org/10.3141%2F2646-08

        In a nutshell, we compute the shortest path between the nodes corresponding to the links to which
        stops were geographically matched, for each pair of identified links.

        We do not consider links that are in perfect sequence, as we found that it introduces severe issues when
        stops are close to intersections without clear upsides.

        When issues are found, we remove the stops in the immediate vicinity of the issue and attempt new
        path finding. The First and last stops/corresponding links are always kept.

        If an error was found, (record for it will show in the log), it is stored within the object.

        """
        if self.__map_matched:
            return
        self.__logger.debug(f"Map-matching pattern ID {self.pattern_id}")

        if not self.__feed.graphs:
            self.__feed.builds_link_graphs_with_broken_stops()

        if not mode_correspondence[self.route_type] in self.__feed.graphs:
            return

        self.__feed.path_store.add_graph(
            self.__feed.graphs[mode_correspondence[self.route_type]], mode_correspondence[self.route_type]
        )

        self.__map_matching_error.clear()
        df = self.__map_matching_complete_path_building()
        if df.shape[0] == 0:
            logger.warning(f"Could not rebuild path for pattern {self.pattern_id}")
            return
        self.full_path = df.link_id.to_list()
        self.fpath_dir = df.dir.to_list()
        self.__assemble__mm_shape(df)
        self.__build_pattern_mapping()
        logger.info(f"Map-matched pattern {self.pattern_id}")

    def __graph_discount(self, connected_stops):
        self.__geolinks = self.__geolinks[self.__geolinks.modes.str.contains(mode_correspondence[self.route_type])]
        link_idx = GeoIndex()
        for _, row in self.__geolinks.iterrows():
            link_idx.insert(feature_id=row.link_id, geometry=row.geometry)
        links = set()
        for stop in connected_stops:
            links.update(list(link_idx.nearest(stop.geo, 3)))
        buffer = self.best_shape().buffer(40)  # type: Polygon

        return [
            lnk
            for lnk in links
            if buffer.contains(self.__geolinks.loc[self.__geolinks.link_id == lnk].geometry.values[0])
        ]

    def __map_matching_complete_path_building(self):
        mode_ = mode_correspondence[self.route_type]
        # We get the graph for our job
        graph = self.__feed.graphs[mode_]
        empty_frame = pd.DataFrame([])

        # We search for disconnected stops:
        candidate_stops = [stop for stop in self.stops]
        stop_node_idxs = [stop.___map_matching_id__[self.route_type] for stop in candidate_stops]

        node0 = graph.network.a_node[~graph.network.a_node.isin(graph.centroids)].min()
        connected_stops = []
        for i, stop in enumerate(candidate_stops):
            node_o = stop.___map_matching_id__[self.route_type]
            logger.debug(f"Computing paths between {node_o} and {node0}")
            res = self.__feed.path_store.get_path_results(node_o, mode_)
            res.update_trace(int(node0))

            # Get skims, as proxy for connectivity, for all stops other than the origin
            other_nodes = stop_node_idxs[:i] + stop_node_idxs[i + 1 :]
            dest_skim = res.skims[other_nodes, 0]
            if dest_skim.min() < 1.0e308:
                candidate_stops = candidate_stops[i:]
                connected_stops = [stop for i, stop in enumerate(candidate_stops[1:]) if dest_skim[i] < 1.0e308]
                connected_stops = [candidate_stops[0]] + connected_stops
                break

        if not connected_stops:
            self.__logger.critical(f"Route completely disconnected. {self.route}/{self.route_id}")
            return empty_frame

        graph.cost = np.array(graph.graph.distance)
        likely_links = self.__graph_discount(connected_stops)
        graph.cost[graph.graph.original_id.abs().isin(likely_links)] *= 0.1

        fstop = connected_stops[0]

        if len(connected_stops) == 1:
            return empty_frame

        if len(connected_stops) == 2:
            nstop = connected_stops[1].___map_matching_id__[self.route_type]
            logger.debug(f"Computing paths between {fstop.___map_matching_id__[self.route_type]} and {nstop}")
            res = self.__feed.path_store.get_path_results(fstop.___map_matching_id__[self.route_type], mode_)
            res.update_trace(int(nstop))
            if res.milepost is None:
                return empty_frame
            pdist = list(res.milepost[1:-1] - res.milepost[:-2])[1:]
            plnks = list(res.path[1:-1])
            pdirecs = list(res.path_link_directions[1:-1])
            return self.__build_path_df(graph, pdirecs, pdist, plnks)

        path_links = []
        path_directions = []
        path_distances = []
        start = fstop.___map_matching_id__[self.route_type]
        for idx, tstop in enumerate(connected_stops[1:]):
            end = tstop.___map_matching_id__[self.route_type]

            not_last = idx + 2 <= len(connected_stops) - 1

            if not_last:
                following_stop = connected_stops[idx + 2]
                n_end = following_stop.___map_matching_id__[self.route_type]
            logger.debug(f"Computing paths between {start} and {end}")
            res = self.__feed.path_store.get_path_results(start, mode_)
            res.update_trace(int(end))
            connection_candidates = graph.network[graph.network.a_node == end].b_node.values
            min_cost = np.inf
            access_node = -1
            follow_val = 0
            for connec in connection_candidates:
                if connec == start:
                    continue
                if not_last:
                    res1 = self.__feed.path_store.get_path_results(connec, mode_)
                    res1.update_trace(int(n_end))
                    if res1.milepost is None:
                        continue
                    follow_val = res1.milepost[-1]
                estimate = follow_val + res.skims[connec, 0]
                if estimate < min_cost:
                    min_cost = estimate
                    access_node = connec
            if access_node >= 0:
                res.update_trace(int(access_node))
                shift = 1 if not_last else 0
                if len(res.path) <= 1 + shift:
                    # Stop connectors only
                    continue

                if not_last:
                    path_links.extend(list(res.path[:-1]))
                    path_directions.extend(list(res.path_link_directions[:-1]))
                    path_distances.extend(list(res.milepost[1:] - res.milepost[:-1])[1:])
                else:
                    path_links.extend(list(res.path[:]))
                    path_directions.extend(list(res.path_link_directions[:]))
                    path_distances.extend(list(res.milepost[1:] - res.milepost[:-1])[:])
            else:
                self.__logger.debug(f"Failed path computation when map-matching {self.pattern_id}")
                return empty_frame
            start = res.path_nodes[-2] if len(res.path_nodes) > 3 else start

        # Connection to the last stop
        return self.__build_path_df(graph, path_directions, path_distances, path_links)

    def __build_path_df(self, graph, path_directions, path_distances, path_links):
        corresp = pd.DataFrame(graph.network[["link_id", "original_id"]])
        if not path_links:
            return pd.DataFrame({"link_id": [], "dir": []})
        result = pd.DataFrame(
            {
                "link_id": path_links[1:],
                "direction": path_directions[1:],
                "sequence": np.arange(len(path_links) - 1),
                "distance": path_distances[1:],
            }
        )
        df = result.merge(corresp, on="link_id", how="left")
        df.sort_values(by=["sequence"], inplace=True)  # We just guarantee that we haven't messed up anything
        df = df[(df.original_id.shift(-1) != df.original_id) | (df.direction.shift(-1) != df.direction)]

        crit1 = df.original_id.shift(1) != df.original_id
        crit2 = df.original_id.shift(-1) != df.original_id
        df = df[(crit1 & crit2) | (df.distance > DEAD_END_RUN)]

        df = df[["original_id", "direction"]]
        df.columns = ["link_id", "dir"]
        df.loc[df.link_id > 0, "dir"] = 0
        df.loc[df.link_id < 0, "dir"] = 1
        df.reset_index(drop=True, inplace=True)
        has_issues = True
        while has_issues:
            # We eliminate multiple backs-and-forth on links
            has_issues = False
            for i in range(0, df.shape[0] - 2):
                if df.loc[i : i + 2, "link_id"].abs().unique().shape[0] == 1:
                    df.drop(index=[i, i + 1], inplace=True)
                    df.reset_index(drop=True, inplace=True)
                    has_issues = True
                    break
        return df

    def __assemble__mm_shape(self, df: pd.DataFrame):
        shape = []  # type: List[Tuple[float, float]]

        for _, rec in df.iterrows():
            line_geo = self.__geolinks.loc[self.__geolinks.link_id == abs(rec.link_id)].geometry.values[0]
            coords = list(line_geo.coords)[::-1] if rec.link_id < 0 else list(line_geo.coords)
            data = coords[1:] if shape else coords
            shape.extend(data)
        self.shape = LineString(shape)

    def get_error(self, what_to_get="culprit") -> Optional[tuple]:
        """
           Returns information on the  area of the network a map-matching error occurred

        Args:

           *what_to_get* (:obj:`str`): The object you want returned. Options are 'culprit' and 'partial_path'

           :return:
        """
        if not self.__map_matching_error:
            self.__logger.debug("No map-matching error recorded for this pattern")
            return None

        if what_to_get not in self.__map_matching_error:
            return None
        return self.__map_matching_error[what_to_get]

    def __build_pattern_mapping(self):
        # We find what is the position along routes that we have for each stop and make sure they are always growing
        self.pattern_mapping = []
        segments = [LineString([pt1, pt2]) for pt1, pt2 in zip(self.shape.coords, self.shape.coords[1:])]
        seg_len = [seg.length for seg in segments]
        distances = []

        min_idx_so_far = 0
        for i, stop in enumerate(self.stops):
            d = [round(stop.geo.distance(line), 1) for line in segments]
            idx = d[min_idx_so_far:].index(min(d[min_idx_so_far:])) + min_idx_so_far
            idx = idx if i > 0 else 0
            projection = segments[idx].project(stop.geo) + sum(seg_len[:idx])
            distances.append(projection)
            min_idx_so_far = idx

        # the list needs to be monotonically increasing
        if not all(x <= y for x, y in zip(distances, distances[1:])):
            for i, (x, y) in enumerate(zip(distances, distances[1:])):
                distances[i + 1] = distances[i] if y < x else distances[i + 1]

            logger.warning(
                f"""Pattern {self.pattern_id} has a map-matching pattern resulting in backwards movement.
                            It was fixed, but it should be checked."""
            )

        distances[-1] = distances[-1] if self.shape.length > distances[-1] else self.shape.length - 0.00001

        # We make sure we don't have projections going beyond the length we will accumulate
        # This is only required because of numerical precision issues
        tot_broken_length = sum(
            [self.__geolinks.loc[self.__geolinks.link_id == abs(x)].geometry.values[0].length for x in self.full_path]
        )
        distances = [min(x, tot_broken_length) for x in distances]
        pid = self.pattern_id
        stop_pos = 0
        cum_dist = 0
        index = 0
        for link_pos, link_id in enumerate([abs(x) for x in self.full_path]):
            direc = self.fpath_dir[link_pos]
            link_geo = self.__geolinks.loc[self.__geolinks.link_id == link_id].geometry.values[0]
            link_geo = link_geo if direc == 0 else LineString(link_geo.coords[::-1])
            if distances[stop_pos] > cum_dist + link_geo.length:
                # If we have no stops falling right in this link
                dt = [pid, index, link_id, direc, link_geo.wkb]
                self.pattern_mapping.append(dt)
                index += 1
                cum_dist += link_geo.length
                continue

            start_point = cum_dist
            end_point = cum_dist + link_geo.length
            while cum_dist + link_geo.length >= distances[stop_pos]:
                milepost = distances[stop_pos]
                wkb = None if stop_pos == 0 else substring(self.shape, start_point, milepost).wkb
                stop = self.stops[stop_pos]
                dt = [pid, index, link_id, direc, wkb]
                self.pattern_mapping.append(dt)
                index += 1

                stop_pos += 1
                if stop_pos >= len(distances):
                    break
                start_point = milepost

            cum_dist += link_geo.length

            if start_point != end_point:
                wkb = substring(self.shape, start_point, end_point).wkb
                self.pattern_mapping.append([pid, index, link_id, direc, wkb])
                index += 1

            if stop_pos >= len(distances):
                break
        self.__compute_match_quality()

    def __compute_match_quality(self):
        dispersion = [(self.shape.distance(stop.geo) * math.pi * 6371000 / 180) for stop in self.stops]
        dispersion = sum([x * x for x in dispersion]) / len(self.stops)
        self.__match_quality = dispersion



"""
 Original Author:  UNKNOWN. COPIED FROM STACKOVERFLOW BUT CAN'T REMEMBER EXACTLY WHERE
 """
import importlib.util as iutil

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import QThread
    from PyQt5.QtCore import pyqtSignal
else:

    class QThread:
        def __init__(self, *arg):
            pass


class WorkerThread(QThread):
    if pyqt:
        jobFinished = pyqtSignal(object)

    def __init__(self, parentThread):
        QThread.__init__(self, parentThread)

    def run(self):
        self.running = True
        success = self.doWork()
        if pyqt:
            self.jobFinished.emit(success)

    def stop(self):
        self.running = False
        pass

    def doWork(self):
        return True

    def cleanUp(self):
        pass



import os
import pathlib
import uuid
import zipfile
from os.path import join, dirname
from tempfile import gettempdir
from unittest import TestCase

import pandas as pd

from aequilibrae import TrafficAssignment, TrafficClass, Graph, Project
from ...data import siouxfalls_project


class TestTrafficAssignmentPathFiles(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        proj_path = os.path.join(gettempdir(), "test_traffic_assignment_path_files" + uuid.uuid4().hex)
        os.mkdir(proj_path)
        zipfile.ZipFile(join(dirname(siouxfalls_project), "sioux_falls_single_class.zip")).extractall(proj_path)
        self.project = Project()
        self.project.open(proj_path)
        self.project.network.build_graphs()
        self.car_graph = self.project.network.graphs["c"]  # type: Graph
        self.car_graph.set_graph("free_flow_time")
        self.car_graph.set_blocked_centroid_flows(False)
        self.matrix = self.project.matrices.get_matrix("demand_omx")
        self.matrix.computational_view()

        self.assignment = TrafficAssignment()
        self.assigclass = TrafficClass("car", self.car_graph, self.matrix)

        self.algorithms = ["msa", "cfw", "bfw", "frank-wolfe"]

    def tearDown(self) -> None:
        self.matrix.close()
        self.project.close()

    def test_set_save_path_files(self):
        self.assignment.set_classes([self.assigclass])
        # make sure default is false
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.save_path_file, False)
        self.assignment.set_save_path_files(True)
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.save_path_file, True)

        # reset for most assignment tests
        self.assignment.set_save_path_files(False)
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.save_path_file, False)

    def test_set_path_file_format(self):
        self.assignment.set_classes([self.assigclass])
        with self.assertRaises(Exception):
            self.assignment.set_path_file_format("shiny_format")
        self.assignment.set_path_file_format("parquet")
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.write_feather, False)
        self.assignment.set_path_file_format("feather")
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.write_feather, True)

    def test_save_path_files(self):
        self.assignment.add_class(self.assigclass)
        self.assignment.set_save_path_files(True)

        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 2
        self.assignment.set_algorithm("msa")
        self.assignment.execute()

        pid = self.assignment.procedure_id
        path_file_dir = pathlib.Path(join(str(self.project.project_base_path), "path_files", pid))
        self.assertTrue(path_file_dir.is_dir())

        # compare everything to reference files. Note that there is no graph simplification happening in SiouxFalls
        # and therefore we compare the files directly, otherwise a translation from the simplified ids to link_ids
        # would need to be performed.
        # Reference files were generated on 12/6/21, any changes to the test project will need to be applied to the
        # reference files. Also, the name given to the traffic class (see setUp above) has to be "car".
        class_id = f"c{self.assigclass.mode}_{self.assigclass.__id__}"
        reference_path_file_dir = pathlib.Path(siouxfalls_project) / "path_files"

        ref_node_correspondence = pd.read_feather(reference_path_file_dir / f"nodes_to_indices_{class_id}.feather")
        node_correspondence = pd.read_feather(path_file_dir / f"nodes_to_indices_{class_id}.feather")
        ref_node_correspondence.node_index = ref_node_correspondence.node_index.astype(
            node_correspondence.node_index.dtype
        )
        self.assertTrue(node_correspondence.equals(ref_node_correspondence))

        ref_correspondence = pd.read_feather(reference_path_file_dir / f"correspondence_{class_id}.feather")
        correspondence = pd.read_feather(path_file_dir / f"correspondence_{class_id}.feather")
        for col in correspondence.columns:
            ref_correspondence[col] = ref_correspondence[col].astype(correspondence[col].dtype)
        self.assertTrue(correspondence.equals(ref_correspondence))

        path_class_id = f"path_{class_id}"
        for i in range(1, self.assignment.max_iter + 1):
            class_dir = path_file_dir / f"iter{i}" / path_class_id
            ref_class_dir = reference_path_file_dir / f"iter{i}" / path_class_id
            for o in self.assigclass.matrix.index:
                o_ind = self.assigclass.graph.compact_nodes_to_indices[o]
                this_o_path_file = pd.read_feather(class_dir / f"o{o_ind}.feather")
                ref_this_o_path_file = pd.read_feather(ref_class_dir / f"o{o_ind}.feather")
                pd.testing.assert_frame_equal(ref_this_o_path_file, this_o_path_file)

                this_o_index_file = pd.read_feather(class_dir / f"o{o_ind}_indexdata.feather")
                ref_this_o_index_file = pd.read_feather(ref_class_dir / f"o{o_ind}_indexdata.feather")
                pd.testing.assert_frame_equal(ref_this_o_index_file, this_o_index_file)



import shapely.wkb
from shapely.geometry import Polygon

from aequilibrae.project.field_editor import FieldEditor


class BasicTable:
    """
    Basic resources used by all subclasses
    """

    def __init__(self, project):
        self.project = project
        self.__table_type__ = ""
        self.conn = project.connect()
        self._curr = self.conn.cursor()

    def extent(self) -> Polygon:
        """Queries the extent of thelayer  included in the model

        Returns:
            *model extent* (:obj:`Polygon`): Shapely polygon with the bounding box of the layer.
        """
        self.__curr.execute(f'Select ST_asBinary(GetLayerExtent("{self.__table_type__}"))')
        poly = shapely.wkb.loads(self.__curr.fetchone()[0])
        return poly

    @property
    def fields(self) -> FieldEditor:
        """Returns a FieldEditor class instance to edit the zones table fields and their metadata"""
        return FieldEditor(self.project, self.__table_type__)

    def refresh_connection(self):
        """Opens a new database connection to avoid thread conflict"""
        self.conn = self.project.connect()
        self.__curr = self.conn.cursor()

    def __copy__(self):
        raise Exception(f"{self.__table_type__} object cannot be copied")

    def __deepcopy__(self, memodict=None):
        raise Exception(f"{self.__table_type__} object cannot be copied")



import importlib.util as iutil
import math
from sqlite3 import Connection as sqlc
from typing import Dict

import numpy as np
import pandas as pd
import shapely.wkb
import shapely.wkt
from shapely.geometry import Polygon
from shapely.ops import unary_union

from aequilibrae.context import get_logger
from aequilibrae.parameters import Parameters
from aequilibrae.project.network import OSMDownloader
from aequilibrae.project.network.gmns_builder import GMNSBuilder
from aequilibrae.project.network.gmns_exporter import GMNSExporter
from aequilibrae.project.network.haversine import haversine
from aequilibrae.project.network.link_types import LinkTypes
from aequilibrae.project.network.links import Links
from aequilibrae.project.network.modes import Modes
from aequilibrae.project.network.nodes import Nodes
from aequilibrae.project.network.osm_builder import OSMBuilder
from aequilibrae.project.network.osm_utils.place_getter import placegetter
from aequilibrae.project.project_creation import req_link_flds, req_node_flds, protected_fields
from aequilibrae.utils import WorkerThread

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import pyqtSignal as SIGNAL


class Network(WorkerThread):
    """
    Network class. Member of an AequilibraE Project
    """

    if pyqt:
        netsignal = SIGNAL(object)

    req_link_flds = req_link_flds
    req_node_flds = req_node_flds
    protected_fields = protected_fields
    link_types: LinkTypes = None

    def __init__(self, project) -> None:
        from aequilibrae.paths import Graph

        WorkerThread.__init__(self, None)
        self.conn = project.conn  # type: sqlc
        self.source = project.source  # type: sqlc
        self.graphs = {}  # type: Dict[Graph]
        self.project = project
        self.logger = project.logger
        self.modes = Modes(self)
        self.link_types = LinkTypes(self)
        self.links = Links(self)
        self.nodes = Nodes(self)

    def skimmable_fields(self):
        """
        Returns a list of all fields that can be skimmed

        :Returns:
            :obj:`list`: List of all fields that can be skimmed
        """
        curr = self.conn.cursor()
        curr.execute("PRAGMA table_info(links);")
        field_names = curr.fetchall()
        ignore_fields = ["ogc_fid", "geometry"] + self.req_link_flds

        skimmable = [
            "INT",
            "INTEGER",
            "TINYINT",
            "SMALLINT",
            "MEDIUMINT",
            "BIGINT",
            "UNSIGNED BIG INT",
            "INT2",
            "INT8",
            "REAL",
            "DOUBLE",
            "DOUBLE PRECISION",
            "FLOAT",
            "DECIMAL",
            "NUMERIC",
        ]
        all_fields = []

        for f in field_names:
            if f[1] in ignore_fields:
                continue
            for i in skimmable:
                if i in f[2].upper():
                    all_fields.append(f[1])
                    break

        all_fields.append("distance")
        real_fields = []
        for f in all_fields:
            if f[-2:] == "ab":
                if f[:-2] + "ba" in all_fields:
                    real_fields.append(f[:-3])
            elif f[-3:] != "_ba":
                real_fields.append(f)

        return real_fields

    def list_modes(self):
        """
        Returns a list of all the modes in this model

        :Returns:
            :obj:`list`: List of all modes
        """
        curr = self.conn.cursor()
        curr.execute("""select mode_id from modes""")
        return [x[0] for x in curr.fetchall()]

    def create_from_osm(
        self,
        west: float = None,
        south: float = None,
        east: float = None,
        north: float = None,
        place_name: str = None,
        modes=["car", "transit", "bicycle", "walk"],
    ) -> None:
        """
        Downloads the network from Open-Street Maps

        :Arguments:
            **west** (:obj:`float`, Optional): West most coordinate of the download bounding box

            **south** (:obj:`float`, Optional): South most coordinate of the download bounding box

            **east** (:obj:`float`, Optional): East most coordinate of the download bounding box

            **place_name** (:obj:`str`, Optional): If not downloading with East-West-North-South boundingbox, this is
            required

            **modes** (:obj:`list`, Optional): List of all modes to be downloaded. Defaults to the modes in the parameter
            file

        .. code-block:: python

            >>> from aequilibrae import Project

            >>> p = Project()
            >>> p.new("/tmp/new_project")

            # We now choose a different overpass endpoint (say a deployment in your local network)
            >>> par = Parameters()
            >>> par.parameters['osm']['overpass_endpoint'] = "http://192.168.1.234:5678/api"

            # Because we have our own server, we can set a bigger area for download (in M2)
            >>> par.parameters['osm']['max_query_area_size'] = 10000000000

            # And have no pause between successive queries
            >>> par.parameters['osm']['sleeptime'] = 0

            # Save the parameters to disk
            >>> par.write_back()

            # Now we can import the network for any place we want
            # p.network.create_from_osm(place_name="my_beautiful_hometown")

            >>> p.close()
        """

        if self.count_links() > 0:
            raise FileExistsError("You can only import an OSM network into a brand new model file")

        curr = self.conn.cursor()
        curr.execute("""ALTER TABLE links ADD COLUMN osm_id integer""")
        curr.execute("""ALTER TABLE nodes ADD COLUMN osm_id integer""")
        self.conn.commit()

        if isinstance(modes, (tuple, list)):
            modes = list(modes)
        elif isinstance(modes, str):
            modes = [modes]
        else:
            raise ValueError("'modes' needs to be string or list/tuple of string")

        if place_name is None:
            if min(east, west) < -180 or max(east, west) > 180 or min(north, south) < -90 or max(north, south) > 90:
                raise ValueError("Coordinates out of bounds")
            bbox = [west, south, east, north]
        else:
            bbox, report = placegetter(place_name)
            west, south, east, north = bbox
            if bbox is None:
                msg = f'We could not find a reference for place name "{place_name}"'
                self.logger.warning(msg)
                return
            for i in report:
                if "PLACE FOUND" in i:
                    self.logger.info(i)

        # Need to compute the size of the bounding box to not exceed it too much
        height = haversine((east + west) / 2, south, (east + west) / 2, north)
        width = haversine(east, (north + south) / 2, west, (north + south) / 2)
        area = height * width

        par = Parameters().parameters["osm"]
        max_query_area_size = par["max_query_area_size"]

        if area < max_query_area_size:
            polygons = [bbox]
        else:
            polygons = []
            parts = math.ceil(area / max_query_area_size)
            horizontal = math.ceil(math.sqrt(parts))
            vertical = math.ceil(parts / horizontal)
            dx = (east - west) / horizontal
            dy = (north - south) / vertical
            for i in range(horizontal):
                xmin = max(-180, west + i * dx)
                xmax = min(180, west + (i + 1) * dx)
                for j in range(vertical):
                    ymin = max(-90, south + j * dy)
                    ymax = min(90, south + (j + 1) * dy)
                    box = [xmin, ymin, xmax, ymax]
                    polygons.append(box)
        self.logger.info("Downloading data")
        self.downloader = OSMDownloader(polygons, modes, logger=self.logger)
        if pyqt:
            self.downloader.downloading.connect(self.signal_handler)

        self.downloader.doWork()

        self.logger.info("Building Network")
        self.builder = OSMBuilder(self.downloader.json, self.source, project=self.project)

        if pyqt:
            self.builder.building.connect(self.signal_handler)
        self.builder.doWork()

        self.logger.info("Network built successfully")

    def create_from_gmns(
        self,
        link_file_path: str,
        node_file_path: str,
        use_group_path: str = None,
        geometry_path: str = None,
        srid: int = 4326,
    ) -> None:
        """
        Creates AequilibraE model from links and nodes in GMNS format.

        :Arguments:
            **link_file_path** (:obj:`str`): Path to a links csv file in GMNS format

            **node_file_path** (:obj:`str`): Path to a nodes csv file in GMNS format

            **use_group_path** (:obj:`str`, Optional): Path to a csv table containing groupings of uses. This helps AequilibraE
            know when a GMNS use is actually a group of other GMNS uses

            **geometry_path** (:obj:`str`, Optional): Path to a csv file containing geometry information for a line object, if not
            specified in the link table

            **srid** (:obj:`int`, Optional): Spatial Reference ID in which the GMNS geometries were created
        """

        gmns_builder = GMNSBuilder(self, link_file_path, node_file_path, use_group_path, geometry_path, srid)
        gmns_builder.doWork()

        self.logger.info("Network built successfully")

    def export_to_gmns(self, path: str):
        """
        Exports AequilibraE network to csv files in GMNS format.

        :Arguments:
            **path** (:obj:`str`): Output folder path.
        """

        gmns_exporter = GMNSExporter(self, path)
        gmns_exporter.doWork()

        self.logger.info("Network exported successfully")

    def signal_handler(self, val):
        if pyqt:
            self.netsignal.emit(val)

    def build_graphs(self, fields: list = None, modes: list = None) -> None:
        """Builds graphs for all modes currently available in the model

        When called, it overwrites all graphs previously created and stored in the networks'
        dictionary of graphs

        :Arguments:
            **fields** (:obj:`list`, optional): When working with very large graphs with large number of fields in the
                                              database, it may be useful to specify which fields to use
            **modes** (:obj:`list`, optional): When working with very large graphs with large number of fields in the
                                              database, it may be useful to generate only those we need

        To use the *fields* parameter, a minimalistic option is the following

        .. code-block:: python

            >>> from aequilibrae import Project

            >>> p = Project.from_path("/tmp/test_project")
            >>> fields = ['distance']
            >>> p.network.build_graphs(fields, modes = ['c', 'w'])

        """
        from aequilibrae.paths import Graph

        curr = self.conn.cursor()

        if fields is None:
            curr.execute("PRAGMA table_info(links);")
            field_names = curr.fetchall()

            ignore_fields = ["ogc_fid", "geometry"]
            all_fields = [f[1] for f in field_names if f[1] not in ignore_fields]
        else:
            fields.extend(["link_id", "a_node", "b_node", "direction", "modes"])
            all_fields = list(set(fields))

        if modes is None:
            modes = curr.execute("select mode_id from modes;").fetchall()
            modes = [m[0] for m in modes]
        elif isinstance(modes, str):
            modes = [modes]

        sql = f"select {','.join(all_fields)} from links"

        df = pd.read_sql(sql, self.conn).fillna(value=np.nan)
        valid_fields = list(df.select_dtypes(np.number).columns) + ["modes"]
        curr.execute("select node_id from nodes where is_centroid=1 order by node_id;")
        centroids = np.array([i[0] for i in curr.fetchall()], np.uint32)

        data = df[valid_fields]
        for m in modes:
            net = pd.DataFrame(data, copy=True)
            net.loc[~net.modes.str.contains(m), "b_node"] = net.loc[~net.modes.str.contains(m), "a_node"]
            g = Graph()
            g.mode = m
            g.network = net
            if centroids.shape[0]:
                g.prepare_graph(centroids)
                g.set_blocked_centroid_flows(True)
            else:
                get_logger().warning("Your graph has no centroids")
            self.graphs[m] = g

    def set_time_field(self, time_field: str) -> None:
        """
        Set the time field for all graphs built in the model

        :Arguments:
            **time_field** (:obj:`str`): Network field with travel time information
        """
        for m, g in self.graphs.items():
            if time_field not in list(g.graph.columns):
                raise ValueError(f"{time_field} not available. Check if you have NULL values in the database")
            g.free_flow_time = time_field
            g.set_graph(time_field)
            self.graphs[m] = g

    def count_links(self) -> int:
        """
        Returns the number of links in the model

        :Returns:
            :obj:`int`: Number of links
        """
        return self.__count_items("link_id", "links", "link_id>=0")

    def count_centroids(self) -> int:
        """
        Returns the number of centroids in the model

        :Returns:
            :obj:`int`: Number of centroids
        """
        return self.__count_items("node_id", "nodes", "is_centroid=1")

    def count_nodes(self) -> int:
        """
        Returns the number of nodes in the model

        :Returns:
            :obj:`int`: Number of nodes
        """
        return self.__count_items("node_id", "nodes", "node_id>=0")

    def extent(self):
        """Queries the extent of the network included in the model

        :Returns:
            **model extent** (:obj:`Polygon`): Shapely polygon with the bounding box of the model network.
        """
        curr = self.conn.cursor()
        curr.execute('Select ST_asBinary(GetLayerExtent("Links"))')
        poly = shapely.wkb.loads(curr.fetchone()[0])
        return poly

    def convex_hull(self) -> Polygon:
        """Queries the model for the convex hull of the entire network

        :Returns:
            **model coverage** (:obj:`Polygon`): Shapely (Multi)polygon of the model network.
        """
        curr = self.conn.cursor()
        curr.execute('Select ST_asBinary("geometry") from Links where ST_Length("geometry") > 0;')
        links = [shapely.wkb.loads(x[0]) for x in curr.fetchall()]
        return unary_union(links).convex_hull

    def refresh_connection(self):
        """Opens a new database connection to avoid thread conflict"""
        self.conn = self.project.connect()

    def __count_items(self, field: str, table: str, condition: str) -> int:
        c = self.conn.execute(f"select count({field}) from {table} where {condition};").fetchone()[0]
        return c



import numpy as np
import multiprocessing as mp
from aequilibrae.matrix.aequilibrae_matrix import AequilibraeMatrix
from aequilibrae.paths.graph import Graph


class SkimResults:
    """
    Network skimming result holder.

    .. code-block:: python

          >>> from aequilibrae import Project
          >>> from aequilibrae.paths.results import SkimResults

          >>> proj = Project.from_path("/tmp/test_project")
          >>> proj.network.build_graphs()

          # Mode c is car in this project
          >>> car_graph = proj.network.graphs['c']

          # minimize travel time
          >>> car_graph.set_graph('free_flow_time')

          # Skims travel time and distance
          >>> car_graph.set_skimming(['free_flow_time', 'distance'])

          >>> res = SkimResults()
          >>> res.prepare(car_graph)

          >>> res.skims.export('/tmp/test_project/matrix.aem')
    """

    def __init__(self):
        self.skims = AequilibraeMatrix()
        self.cores = mp.cpu_count()

        self.links = -1
        self.nodes = -1
        self.zones = -1
        self.num_skims = -1
        self.__graph_id__ = None
        self.graph = Graph()

    def prepare(self, graph: Graph):
        """
        Prepares the object with dimensions corresponding to the graph objects

        :Arguments:
            **graph** (:obj:`Graph`): Needs to have been set with number of centroids and list of skims (if any)
        """

        if not graph.cost_field:
            raise Exception('Cost field needs to be set for computation. use graph.set_graph("your_cost_field")')

        self.nodes = graph.compact_num_nodes + 1
        self.zones = graph.num_zones
        self.links = graph.compact_num_links + 1
        self.num_skims = len(graph.skim_fields)

        self.skims = AequilibraeMatrix()
        self.skims.create_empty(
            file_name=AequilibraeMatrix().random_name(), zones=self.zones, matrix_names=graph.skim_fields
        )
        self.skims.index[:] = graph.centroids[:]
        self.skims.computational_view(core_list=self.skims.names)
        self.skims.matrix_view = self.skims.matrix_view.reshape(self.zones, self.zones, self.num_skims)
        self.__graph_id__ = graph.__id__
        self.graph = graph

    def set_cores(self, cores: int) -> None:
        """
        Sets number of cores (threads) to be used in computation

        Value of zero sets number of threads to all available in the system, while negative values indicate the number
        of threads to be left out of the computational effort.

        Resulting number of cores will be adjusted to a minimum of zero or the maximum available in the system if the
        inputs result in values outside those limits

        :Arguments:
            **cores** (:obj:`int`): Number of cores to be used in computation
        """

        if isinstance(cores, int):
            if cores < 0:
                self.cores = max(1, mp.cpu_count() + cores)
            if cores == 0:
                self.cores = mp.cpu_count()
            elif cores > 0:
                cores = min(mp.cpu_count(), cores)
                if self.cores != cores:
                    self.cores = cores
        else:
            raise ValueError("Number of cores needs to be an integer")



import gc
import importlib.util as iutil
import sqlite3
import string
from typing import List

import numpy as np
import pandas as pd

from aequilibrae.context import get_active_project
from aequilibrae.parameters import Parameters
from aequilibrae.project.network.link_types import LinkTypes
from aequilibrae.utils.spatialite_utils import connect_spatialite
from .haversine import haversine
from ...utils import WorkerThread

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import pyqtSignal

spec = iutil.find_spec("qgis")
isqgis = spec is not None
if isqgis:
    import qgis


class OSMBuilder(WorkerThread):
    if pyqt:
        building = pyqtSignal(object)

    def __init__(self, osm_items: List, path: str, node_start=10000, project=None) -> None:
        WorkerThread.__init__(self, None)
        self.project = project or get_active_project()
        self.logger = self.project.logger
        self.osm_items = osm_items
        self.path = path
        self.conn = None
        self.node_start = node_start
        self.__link_types = None  # type: LinkTypes
        self.report = []
        self.__model_link_types = []
        self.__model_link_type_ids = []
        self.__link_type_quick_reference = {}
        self.nodes = {}
        self.node_df = []
        self.links = {}
        self.insert_qry = """INSERT INTO {} ({}, geometry) VALUES({}, GeomFromText(?, 4326))"""

    def __emit_all(self, *args):
        if pyqt:
            self.building.emit(*args)

    def doWork(self):
        self.conn = connect_spatialite(self.path)
        self.curr = self.conn.cursor()
        self.__worksetup()
        node_count = self.data_structures()
        self.importing_links(node_count)
        self.__emit_all(["finished_threaded_procedure", 0])

    def data_structures(self):
        self.logger.info("Separating nodes and links")
        self.__emit_all(["text", "Separating nodes and links"])
        self.__emit_all(["maxValue", len(self.osm_items)])

        alinks = []
        n = []
        tot_items = len(self.osm_items)
        # When downloading data for entire countries, memory consumption can be quite intensive
        # So we get rid of everything we don't need
        for i in range(tot_items, 0, -1):
            item = self.osm_items.pop(-1)
            if item["type"] == "way":
                alinks.append(item)
            elif item["type"] == "node":
                n.append(item)
            self.__emit_all(["Value", tot_items - i])
        gc.collect()

        self.logger.info("Setting data structures for nodes")
        self.__emit_all(["text", "Setting data structures for nodes"])
        self.__emit_all(["maxValue", len(n)])

        self.node_df = []
        for i, node in enumerate(n):
            nid = node.pop("id")
            _ = node.pop("type")
            node["node_id"] = i + self.node_start
            self.nodes[nid] = node
            self.node_df.append([node["node_id"], nid, node["lon"], node["lat"]])
            self.__emit_all(["Value", i])
        del n
        self.node_df = (
            pd.DataFrame(self.node_df, columns=["A", "B", "C", "D"])
            .drop_duplicates(subset=["C", "D"])
            .to_records(index=False)
        )

        self.logger.info("Setting data structures for links")
        self.__emit_all(["text", "Setting data structures for links"])
        self.__emit_all(["maxValue", len(alinks)])

        all_nodes = []
        for i, link in enumerate(alinks):
            osm_id = link.pop("id")
            _ = link.pop("type")
            all_nodes.extend(link["nodes"])
            self.links[osm_id] = link
            self.__emit_all(["Value", i])
        del alinks

        self.logger.info("Finalizing data structures")
        self.__emit_all(["text", "Finalizing data structures"])

        node_count = self.unique_count(np.array(all_nodes))

        return node_count

    def importing_links(self, node_count):
        node_ids = {}

        vars = {}
        vars["link_id"] = 1
        table = "links"
        fields = self.get_link_fields()
        self.__update_table_structure()
        field_names = ",".join(fields)

        self.logger.info("Adding network nodes")
        self.__emit_all(["text", "Adding network nodes"])
        sql = "insert into nodes(node_id, is_centroid, osm_id, geometry) Values(?, 0, ?, MakePoint(?,?, 4326))"
        self.conn.executemany(sql, self.node_df)
        self.conn.commit()
        del self.node_df

        self.logger.info("Adding network links")
        self.__emit_all(["text", "Adding network links"])
        L = len(list(self.links.keys()))
        self.__emit_all(["maxValue", L])

        counter = 0
        mode_codes, not_found_tags = self.modes_per_link_type()
        owf, twf = self.field_osm_source()
        all_attrs = []
        all_osm_ids = list(self.links.keys())
        for osm_id in all_osm_ids:
            link = self.links.pop(osm_id)
            self.__emit_all(["Value", counter])
            counter += 1
            if counter % 1000 == 0:
                self.logger.info(f"Creating segments from {counter:,} out of {L:,} OSM link objects")
            vars["osm_id"] = osm_id
            vars["link_type"] = "default"
            linknodes = link["nodes"]
            linktags = link["tags"]

            indices = np.searchsorted(node_count[:, 0], linknodes)
            nodedegree = node_count[indices, 1]

            # Makes sure that beginning and end are end nodes for a link
            nodedegree[0] = 2
            nodedegree[-1] = 2

            intersections = np.where(nodedegree > 1)[0]
            segments = intersections.shape[0] - 1

            # Attributes that are common to all individual links/segments
            vars["direction"] = (linktags.get("oneway") == "yes") * 1

            for k, v in owf.items():
                vars[k] = linktags.get(v)

            for k, v in twf.items():
                val = linktags.get(v["osm_source"])
                if vars["direction"] == 0:
                    for d1, d2 in [("ab", "forward"), ("ba", "backward")]:
                        vars[f"{k}_{d1}"] = self.__get_link_property(d2, val, linktags, v)
                elif vars["direction"] == -1:
                    vars[f"{k}_ba"] = linktags.get(f"{v['osm_source']}:{'backward'}", val)
                elif vars["direction"] == 1:
                    vars[f"{k}_ab"] = linktags.get(f"{v['osm_source']}:{'forward'}", val)

            vars["modes"] = mode_codes.get(linktags.get("highway"), not_found_tags)

            vars["link_type"] = self.__link_type_quick_reference.get(
                vars["link_type"].lower(), self.__repair_link_type(vars["link_type"])
            )

            if len(vars["modes"]) > 0:
                for i in range(segments):
                    attributes = self.__build_link_data(vars, intersections, i, linknodes, node_ids, fields)
                    all_attrs.append(attributes)
                    vars["link_id"] += 1

            self.__emit_all(["text", f"{counter:,} of {L:,} super links added"])
            self.links[osm_id] = []
        sql = self.insert_qry.format(table, field_names, ",".join(["?"] * (len(all_attrs[0]) - 1)))
        self.logger.info("Adding network links")
        self.__emit_all(["text", "Adding network links"])
        try:
            self.curr.executemany(sql, all_attrs)
        except Exception as e:
            self.logger.error("error when inserting link {}. Error {}".format(all_attrs[0], e.args))
            self.logger.error(sql)
            raise e

        self.conn.commit()
        self.curr.close()

    def __worksetup(self):
        self.__link_types = self.project.network.link_types
        lts = self.__link_types.all_types()
        for lt_id, lt in lts.items():
            self.__model_link_types.append(lt.link_type)
            self.__model_link_type_ids.append(lt_id)

    def __update_table_structure(self):
        curr = self.conn.cursor()
        curr.execute("pragma table_info(Links)")
        structure = curr.fetchall()
        has_fields = [x[1].lower() for x in structure]
        fields = [field.lower() for field in self.get_link_fields()] + ["osm_id"]
        for field in [f for f in fields if f not in has_fields]:
            ltype = self.get_link_field_type(field).upper()
            curr.execute(f"Alter table Links add column {field} {ltype}")
        self.conn.commit()

    def __build_link_data(self, vars, intersections, i, linknodes, node_ids, fields):
        ii = intersections[i]
        jj = intersections[i + 1]
        all_nodes = [linknodes[x] for x in range(ii, jj + 1)]

        vars["a_node"] = node_ids.get(linknodes[ii], self.node_start)
        if vars["a_node"] == self.node_start:
            node_ids[linknodes[ii]] = vars["a_node"]
            self.node_start += 1

        vars["b_node"] = node_ids.get(linknodes[jj], self.node_start)
        if vars["b_node"] == self.node_start:
            node_ids[linknodes[jj]] = vars["b_node"]
            self.node_start += 1

        vars["distance"] = sum(
            [
                haversine(self.nodes[x]["lon"], self.nodes[x]["lat"], self.nodes[y]["lon"], self.nodes[y]["lat"])
                for x, y in zip(all_nodes[1:], all_nodes[:-1])
            ]
        )

        geometry = ["{} {}".format(self.nodes[x]["lon"], self.nodes[x]["lat"]) for x in all_nodes]
        geometry = "LINESTRING ({})".format(", ".join(geometry))

        attributes = [vars.get(x) for x in fields]
        attributes.append(geometry)
        return attributes

    def __repair_link_type(self, link_type: str) -> str:
        original_link_type = link_type
        link_type = "".join([x for x in link_type if x in string.ascii_letters + "_"]).lower()

        split = link_type.split("_")
        for i, piece in enumerate(split[1:]):
            if piece in ["link", "segment", "stretch"]:
                link_type = "_".join(split[0 : i + 1])

        if len(link_type) == 0:
            link_type = "empty"

        if len(self.__model_link_type_ids) >= 51 and link_type not in self.__model_link_types:
            link_type = "aggregate_link_type"

        if link_type in self.__model_link_types:
            lt = self.__link_types.get_by_name(link_type)
            if original_link_type not in lt.description:
                lt.description += f", {original_link_type}"
                lt.save()
            self.__link_type_quick_reference[original_link_type.lower()] = link_type
            return link_type

        letter = link_type[0]
        if letter in self.__model_link_type_ids:
            letter = letter.upper()
            if letter in self.__model_link_type_ids:
                for letter in string.ascii_letters:
                    if letter not in self.__model_link_type_ids:
                        break
        lt = self.__link_types.new(letter)
        lt.link_type = link_type
        lt.description = f"Link types from Open Street Maps: {original_link_type}"
        lt.save()
        self.__model_link_types.append(link_type)
        self.__model_link_type_ids.append(letter)
        self.__link_type_quick_reference[original_link_type.lower()] = link_type
        return link_type

    def __get_link_property(self, d2, val, linktags, v):
        vald = linktags.get(f'{v["osm_source"]}:{d2}', val)
        if vald is None:
            return vald

        if vald.isdigit():
            if vald == val and v["osm_behaviour"] == "divide":
                vald = float(val) / 2
        return vald

    @staticmethod
    def unique_count(a):
        # From: https://stackoverflow.com/a/21124789/1480643
        unique, inverse = np.unique(a, return_inverse=True)
        count = np.zeros(len(unique), int)
        np.add.at(count, inverse, 1)
        return np.vstack((unique, count)).T

    @staticmethod
    def get_link_fields():
        p = Parameters()
        fields = p.parameters["network"]["links"]["fields"]
        owf = [list(x.keys())[0] for x in fields["one-way"]]

        twf1 = ["{}_ab".format(list(x.keys())[0]) for x in fields["two-way"]]
        twf2 = ["{}_ba".format(list(x.keys())[0]) for x in fields["two-way"]]

        return owf + twf1 + twf2 + ["osm_id"]

    @staticmethod
    def get_link_field_type(field_name):
        p = Parameters()
        fields = p.parameters["network"]["links"]["fields"]

        if field_name[-3:].lower() in ["_ab", "_ba"]:
            field_name = field_name[:-3]
            for tp in fields["two-way"]:
                if field_name in tp:
                    return tp[field_name]["type"]
        else:
            for tp in fields["one-way"]:
                if field_name in tp:
                    return tp[field_name]["type"]

    @staticmethod
    def field_osm_source():
        p = Parameters()
        fields = p.parameters["network"]["links"]["fields"]

        owf = {
            list(x.keys())[0]: x[list(x.keys())[0]]["osm_source"]
            for x in fields["one-way"]
            if "osm_source" in x[list(x.keys())[0]]
        }

        twf = {}
        for x in fields["two-way"]:
            if "osm_source" in x[list(x.keys())[0]]:
                twf[list(x.keys())[0]] = {
                    "osm_source": x[list(x.keys())[0]]["osm_source"],
                    "osm_behaviour": x[list(x.keys())[0]]["osm_behaviour"],
                }
        return owf, twf

    def modes_per_link_type(self):
        p = Parameters()
        modes = p.parameters["network"]["osm"]["modes"]

        cursor = self.conn.cursor()
        cursor.execute("SELECT mode_name, mode_id from modes")
        mode_codes = cursor.fetchall()
        mode_codes = {p[0]: p[1] for p in mode_codes}

        type_list = {}
        notfound = ""
        for mode, val in modes.items():
            all_types = val["link_types"]
            md = mode_codes[mode]
            for tp in all_types:
                type_list[tp] = "{}{}".format(type_list.get(tp, ""), md)
            if val["unknown_tags"]:
                notfound += md

        type_list = {k: "".join(set(v)) for k, v in type_list.items()}

        return type_list, "{}".format(notfound)

    @staticmethod
    def get_node_fields():
        p = Parameters()
        fields = p.parameters["network"]["nodes"]["fields"]
        fields = [list(x.keys())[0] for x in fields]
        return fields + ["osm_id"]
